# !/usr/bin/python
# coding=utf-8
from typing import Union, List, Optional, Dict, Tuple

# from this package:
from pythontk import core_utils
from pythontk import iter_utils
from pythontk import file_utils


class StrUtils(core_utils.CoreUtils):
    """ """

    @staticmethod
    def sanitize(
        text: Union[str, List[str]],
        replacement_char: str = "_",
        char_map: Optional[Dict[str, str]] = None,
        preserve_trailing: bool = False,
        preserve_case: bool = False,
        allow_consecutive: bool = False,
        return_original: bool = False,  # Optionally return original string(s)
    ) -> Union[str, Tuple[str, str], List[str], List[Tuple[str, str]]]:
        """Sanitizes a string or a list of strings by replacing invalid characters.

        Returns:
            (obj/list) dependant on flags.
        """
        import re

        def sanitize_single(text: str) -> Union[str, Tuple[str, str]]:
            original_text = text
            txt = text if preserve_case else text.lower()

            # Apply character mappings if provided
            if char_map:
                for char, replacement in char_map.items():
                    txt = txt.replace(char, replacement)

            # Replace all non-alphanumeric characters
            sanitized_text = re.sub(
                r"[^a-z0-9_]" if not preserve_case else r"[^A-Za-z0-9_]",
                replacement_char,
                txt,
            )

            # Collapse consecutive replacement characters if allow_consecutive is False
            if not allow_consecutive:
                sanitized_text = re.sub(
                    f"{replacement_char}+", replacement_char, sanitized_text
                )

            # Optionally remove trailing illegal characters if preserve_trailing is False
            if not preserve_trailing:
                sanitized_text = re.sub(f"{replacement_char}+$", "", sanitized_text)

            return (
                (sanitized_text, original_text) if return_original else sanitized_text
            )

        # Ensure the input is always iterable using the make_iterable method
        iterable_text = iter_utils.IterUtils.make_iterable(text)

        # Sanitize each item in the iterable
        sanitized_list = [sanitize_single(t) for t in iterable_text]

        # Return the appropriate format using format_return
        return core_utils.CoreUtils.format_return(sanitized_list, orig=text)

    @staticmethod
    @core_utils.CoreUtils.listify(threading=True)
    def set_case(string, case="title"):
        """Format the given string(s) in the given case.

        Parameters:
            string (str/list): The string(s) to format.
            case (str): The desired return case. Accepts all python case operators.
                    valid: 'upper', 'lower', 'capitalize' (default), 'swapcase', 'title', 'pascal', 'camel', None.
        Returns:
            (str/list) List if 'string' given as list.
        """
        if (not string) or (not isinstance(string, str)):
            return ""

        if case == "pascal":
            return string[:1].capitalize() + string[1:]  # capitalize the first letter.

        elif case == "camel":
            return string[0].lower() + string[1:]  # lowercase the first letter.

        else:
            try:
                return getattr(string, case)()

            except AttributeError:  # return the original string.
                return string

    @staticmethod
    def get_mangled_name(class_input, attribute_name):
        """Returns the mangled name for a private attribute of a class.

        Parameters:
            class_input (str/type/instance): The class name as a string, the class itself or an instance of the class.
            attribute_name (str): The original name of the attribute.

        Returns:
            str: The mangled name of the attribute.

        Raises:
            TypeError: If class_input is not a string, a type, or an instance of a class, or if attribute_name is not a string.
            ValueError: If attribute_name does not start with double underscore.

        Example:
            get_mangled_name("MyClass", "__attribute") -> "_MyClass__attribute"
            get_mangled_name(MyClass, "__attribute") -> "_MyClass__attribute"
            get_mangled_name(MyClass(), "__attribute") -> "_MyClass__attribute"
        """
        if not isinstance(attribute_name, str):
            raise TypeError("attribute_name must be a string")
        if not attribute_name.startswith("__"):
            raise ValueError("attribute_name must start with double underscore")

        if isinstance(class_input, str):
            class_name = class_input
        elif isinstance(class_input, type):
            class_name = class_input.__name__
        elif hasattr(class_input, "__class__"):
            class_name = class_input.__class__.__name__
        else:
            raise TypeError(
                "class_input must be a string, a type, or an instance of a class"
            )

        return f"_{class_name}{attribute_name}"

    @staticmethod
    def get_text_between_delimiters(string, start_delim, end_delim, as_string=False):
        """Get any text between the specified start and end delimiters in the given string. The text can be returned as a
        generator (default behavior) or as a single concatenated string if `as_string` is set to True.

        Parameters:
            string (str): The input string to search for matches.
            start_delim (str): The starting delimiter to search for.
            end_delim (str): The ending delimiter to search for.
            as_string (bool, optional): If True, the function returns a single concatenated string of all matches.
                                                                     If False (default), the function returns a generator that yields each match.

        Returns:
            If as_string is False (default): A generator that yields all matches found in the input string.
            If as_string is True: A single concatenated string containing all matches found in the input string.

        Example:
            input_string = "Here is the <!-- start -->first match<!-- end --> and here is the <!-- start -->second match<!-- end -->"

            # Get the matches as a generator (default behavior)
            matches_generator = get_text_between_delimiters(input_string, '<!-- start -->', '<!-- end -->')
            for match in matches_generator:
                    print(match)  # Output: first match (first iteration), second match (second iteration)

            # Get the matches as a single string
            matches_string = get_text_between_delimiters(input_string, '<!-- start -->', '<!-- end -->', as_string=True)
            print(matches_string)  # Output: "first match second match"
        """
        import re

        def extract_matches(string, start_delim, end_delim, start_index=0):
            pattern = re.compile(
                f"{re.escape(start_delim)}(.*?){re.escape(end_delim)}", re.DOTALL
            )
            match = pattern.search(string, start_index)
            if match:
                yield match.group(1).strip()
                yield from extract_matches(string, start_delim, end_delim, match.end())

        if as_string:
            matches = list(extract_matches(string, start_delim, end_delim))
            return " ".join(matches)
        else:
            return extract_matches(string, start_delim, end_delim)

    @staticmethod
    def get_matching_hierarchy_items(
        hierarchy_items,
        target,
        upstream=False,
        exact=False,
        downstream=False,
        reverse=False,
        delimiters="|",
    ):
        """Find the closest match(es) for a given 'target' string in a list of hierarchical strings.

        Parameters:
            hierarchy_items (list): A list of strings representing hierarchical items.
            target (str): A string representing the hierarchical item to find a match for.
            upstream (bool, optional): If True, returns items that are one level up in the hierarchy. Default is False.
            exact (bool, optional): If True, returns only items that are an exact match. Default is False.
            downstream (bool, optional): If True, returns items that are one level down in the hierarchy. Default is False.
            reverse (bool, optional): Reverse the result. Default is False.
            delimiters (str/list, optional): A string containing all characters that can act as delimiters in the hierarchy. Default is "|".

        Returns:
            list: A list of matching items ordered by length.

        Example:
            hierarchy_items = [
                "polygons|mesh|submenu",
                "polygons|submenu",
                "polygons",
                "polygons|mesh",
                "polygons|face",
                "polygons|mesh|other",
            ]

            target = "polygons.mesh"
            get_matching_hierarchy_items(hierarchy_items, target, upstream=True) -> ['polygons']
            get_matching_hierarchy_items(hierarchy_items, target, downstream=True) -> ['polygons|mesh|submenu', 'polygons|mesh|other']
            get_matching_hierarchy_items(hierarchy_items, target, exact=True) -> ['polygons|mesh']
        """
        import re

        pattern = "|".join(
            re.escape(d) for d in iter_utils.IterUtils.make_iterable(delimiters)
        )
        target_parts = re.split(pattern, target)

        def match_hierarchy(item_parts):
            return all(p1 == p2 for p1, p2 in zip(item_parts, target_parts))

        def is_upstream(item_parts):
            return len(item_parts) < len(target_parts)

        def is_downstream(item_parts):
            return len(item_parts) > len(target_parts)

        def filter_items(item):
            item_parts = re.split("|".join(re.escape(d) for d in delimiters), item)

            if exact and item == target:
                return True
            if (
                upstream
                and match_hierarchy(item_parts)
                and is_upstream(item_parts)
                and set(item_parts).issubset(set(target_parts))
            ):
                return True
            if downstream and match_hierarchy(item_parts) and is_downstream(item_parts):
                return True
            return False

        matches = [item for item in hierarchy_items if filter_items(item)]
        return sorted(matches, key=lambda x: len(x), reverse=reverse)

    @staticmethod
    @core_utils.CoreUtils.listify(threading=True)
    def split_at_chars(string, chars="|", occurrence=-1):
        """Split a string containing the given chars at the given occurrence and return
        a two element tuple containing both halves.

        Parameters:
            strings (str/list): The string(s) to operate on.
            chars (str): The chars to split at.
            occurrence (int): The occurrence of the pipe to split at from left.
                    ex. -1 would split at the last occurrence. 0 would split at the first.
                        If the occurrence is out of range, the full string will be
                        returned as: ('original string', '')
        Returns:
            (tuple)(list) two element tuple, or list of two element tuples if multiple strings given.

        Example:
            split_at_chars(['str|ing', 'string']) returns: [('str', 'ing'), ('string', '')]
        """
        split = string.split(chars)

        try:
            s2 = "".join(split[occurrence])
            if chars in string:
                s1 = chars.join(split[:occurrence])
                return (s1, s2)
            else:
                return (s2, "")
        except IndexError:
            return (string, "")

    @classmethod
    def insert(cls, src, ins, at, occurrence=1, before=False):
        """Insert character(s) into a string at a given location.
        if the character doesn't exist, the original string will be returned.

        Parameters:
            src (str): The source string.
            ins (str): The character(s) to insert.
            at (str)(int): The index or char(s) to insert at.
            occurrence (int): Specify which occurrence to insert at.
                        Valid only when 'at' is given as a string.
                        default: The first occurrence.
                        (A value of -1 would insert at the last occurrence)
            before (bool): Specify inserting before or after. default: after
                        Valid only when 'at' is given as a string.
        Returns:
            (str)
        """
        try:
            return "".join((src[:at], str(ins), src[at:]))

        except TypeError:
            # if 'occurrance' is a negative value, search from the right.
            if occurrence < 0:
                i = src.replace(at, " " * len(at), occurrence - 1).rfind(at)
            else:
                i = src.replace(at, " " * len(at), occurrence - 1).find(at)
            return (
                cls.insert(src, str(ins), i if before else i + len(at))
                if i != -1
                else src
            )

    @staticmethod
    def rreplace(string, old, new="", count=None):
        """Replace occurrances in a string from right to left.
        The number of occurrances replaced can be limited by using the 'count' argument.

        Parameters:
            string (str):
            old (str):
            new (str)(int):
            count (int):

        Returns:
            (str)
        """
        if not string or not isinstance(string, str):
            return string

        if count is not None:
            return str(new).join(string.rsplit(old, count))
        else:
            return str(new).join(string.rsplit(old))

    @staticmethod
    @core_utils.CoreUtils.listify(threading=True)
    def truncate(string, length=75, beginning=True, insert=".."):
        """Shorten the given string to the given length.
        An ellipsis will be added to the section trimmed.

        Parameters:
            string (str): The string to truncate.
            length (int): The maximum allowed length before trunicating.
            beginning (bool): Trim starting chars, else; ending.
            insert (str): Chars to add at the trimmed area. (default: ellipsis)

        Returns:
            (str)

        Example:
            truncate('12345678', 4) #returns: '..5678'
        """
        if not string or not isinstance(string, str):
            return string

        if len(string) > length:
            if beginning:  # trim starting chars.
                string = insert + string[-length:]
            else:  # trim ending chars.
                string = string[:length] + insert
        return string

    @staticmethod
    def get_trailing_integers(string, inc=0, as_string=False):
        """Returns any integers from the end of the given string.

        Parameters:
            inc (int): Increment by a step amount. (default: 0)
                    0 does not increment and returns the original number.
            as_string (bool): Return the integers as a string instead of integers.

        Returns:
            (int)

        Example:
            get_trailing_integers('p001Cube1', inc=1) #returns: 2
        """
        import re

        if not string or not isinstance(string, str):
            return string

        m = re.findall(r"\d+\s*$", string)
        result = int(m[0]) + inc if m else None

        if as_string:
            return str(result)
        return result

    @staticmethod
    def find_str(find, strings, regex=False, ignore_case=False):
        """Filter for elements that containing the given string in a list of strings.

        Parameters:
            find (str): The search string. An asterisk denotes startswith*, *endswith, *contains*, and multiple search strings can be separated by pipe chars.
                    wildcards:
                        *chars* - string contains chars.
                        *chars - string endswith chars.
                        chars* - string startswith chars.
                        chars1|chars2 - string matches any of.  can be used in conjuction with other modifiers.
                    regular expressions (if regex True):
                        (.) match any char. ex. re.match('1..', '1111') #returns the regex object <111>
                        (^) match start. ex. re.match('^11', '011') #returns None
                        ($) match end. ex. re.match('11$', '011') #returns the regex object <11>
                        (|) or. ex. re.match('1|0', '011') #returns the regex object <0>
                        (\A,\Z) beginning of a string and end of a string. ex. re.match(r'\A011\Z', '011') #
                        (\b) empty string. (\B matches the empty string anywhere else). ex. re.match(r'\b(011)\b', '011 011 011') #
            strings (list): The string list to search.
            regex (bool): Use regular expressions instead of wildcards.
            ignore_case (bool): Search case insensitive.

        Returns:
            (list)

        Example:
            lst = ['invertVertexWeights', 'keepCreaseEdgeWeight', 'keepBorder', 'keepBorderWeight', 'keepColorBorder', 'keepColorBorderWeight']
            find_str('*Weight*', lst) #find any element that contains the string 'Weight'.
            find_str('Weight$|Weights$', lst, regex=True) #find any element that endswith 'Weight' or 'Weights'.
        """
        # Filter out non-string values
        strings = [s for s in strings if isinstance(s, str)]

        if not find:  # Handle empty search string
            return []

        if regex:
            import re

            try:
                if ignore_case:
                    return [i for i in strings if re.search(find, i, re.IGNORECASE)]
                return [i for i in strings if re.search(find, i)]
            except Exception as e:
                print(f"# Error find_str: in {find}: {e}. #")
                return []

        result = []
        find_parts = find.split("|")  # Split multiple search terms

        for w in find_parts:
            w_ = w.strip("*").rstrip("*")  # Remove wildcard modifiers

            # Apply wildcard search with case-insensitivity while maintaining order
            for i in strings:
                check = i.lower() if ignore_case else i
                term = w_.lower() if ignore_case else w_

                if w.startswith("*") and w.endswith("*"):  # Contains
                    if term in check and i not in result:
                        result.append(i)

                elif w.startswith("*"):  # Endswith
                    if check.endswith(term) and i not in result:
                        result.append(i)

                elif w.endswith("*"):  # Startswith
                    if check.startswith(term) and i not in result:
                        result.append(i)

                else:  # Exact match
                    if check == term and i not in result:
                        result.append(i)

        return result  # Order is now preserved

    @classmethod
    def find_str_and_format(
        cls,
        strings,
        to,
        fltr="",
        regex=False,
        ignore_case=False,
        return_orig_strings=False,
    ):
        """Expanding on the 'find_str' function: Find matches of a string in a list of strings and re-format them.

        Parameters:
            strings (list): A list of string objects to search.
            to (str): An optional asterisk modifier can be used for formatting. An empty string will attempt to remove the part of the string designated in the from argument.
                    "" - (empty string) - strip chars.
                    *chars* - replace only.
                    *chars - replace suffix.
                    **chars - append suffix.
                    chars* - replace prefix.
                    chars** - append prefix.
            fltr (str): See the 'find_str' function's 'fltr' parameter for documentation.
            regex (bool): Use regular expressions instead of wildcards for the 'find' argument.
            ignore_case (bool): Ignore case when searching. Applies only to the 'fltr' parameter's search.
            return_orig_strings (bool): Return the old names as well as the new.

        Returns:
            (list) if return_orig_strings: list of two element tuples containing the original and modified string pairs. [('frm','to')]
                    else: a list of just the new names.
        """
        import re

        # Filter out non-string values
        strings = [s for s in strings if isinstance(s, str)]

        # if 'fltr' is not an empty string; fltr 'strings' for matches using 'fltr'.
        if fltr:
            strings = cls.find_str(fltr, strings, regex=regex, ignore_case=ignore_case)

        # re.sub('[^A-Za-z0-9_:]+', '', fltr) #strip any special chars other than '_'.
        frm_ = fltr.strip("*").rstrip("*")
        # remove any modifiers from the left and right end chars.
        to_ = to.strip("*").rstrip("*")

        result = []
        for orig_str in strings:
            # modifiers
            if to.startswith("*") and to.endswith("*"):  # replace chars
                if frm_:  # Only if frm_ is not empty
                    if ignore_case and not regex:
                        # remove frm_ from the string (case in-sensitive).
                        s = re.sub(re.escape(frm_), to_, orig_str, flags=re.IGNORECASE)
                    else:
                        s = orig_str.replace(frm_, to_)
                else:
                    s = orig_str  # No change if no filter pattern

            elif to.startswith("**"):  # append suffix
                s = orig_str + to_

            elif to.startswith("*"):  # replace suffix
                if frm_:  # Only if frm_ is not empty
                    if ignore_case and not regex:
                        # get the starting index of 'frm_'.
                        try:
                            index = re.search(
                                re.escape(frm_), orig_str, flags=re.IGNORECASE
                            ).start()
                            s = orig_str[:index] + to_
                        except (AttributeError, re.error):
                            s = orig_str + to_  # Fallback: append if pattern not found
                    else:
                        try:
                            s = orig_str.split(frm_)[0] + to_
                        except ValueError:  # empty separator
                            s = orig_str + to_  # Fallback: append
                else:
                    s = orig_str + to_  # Append if no filter pattern

            elif to.endswith("**"):  # append prefix
                s = to_ + orig_str

            elif to.endswith("*"):  # replace prefix
                if frm_:  # Only if frm_ is not empty
                    if ignore_case and not regex:
                        # get the ending index of 'frm_'.
                        try:
                            index = re.search(
                                re.escape(frm_), orig_str, flags=re.IGNORECASE
                            ).end()
                            s = to_ + orig_str[index:]
                        except (AttributeError, re.error):
                            s = to_ + orig_str  # Fallback: prepend if pattern not found
                    else:
                        try:
                            parts = orig_str.split(frm_)
                            if len(parts) > 1:
                                s = to_ + frm_ + orig_str.split(frm_)[-1]
                            else:
                                s = to_ + orig_str
                        except ValueError:  # empty separator
                            s = to_ + orig_str  # Fallback: prepend
                else:
                    s = to_ + orig_str  # Prepend if no filter pattern

            elif not to_:  # if 'to_' is an empty string:
                if frm_:  # Only process if there's something to remove
                    if fltr.endswith("*") and not fltr.startswith(
                        "*"
                    ):  # strip only beginning chars.
                        if ignore_case and not regex:
                            # remove the first instance of frm_ from the string (case in-sensitive).
                            s = re.sub(
                                re.escape(frm_), "", orig_str, 1, flags=re.IGNORECASE
                            )
                        else:
                            # remove first instance of frm_ from the string.
                            s = orig_str.replace(frm_, "", 1)

                    elif fltr.startswith("*") and not fltr.endswith(
                        "*"
                    ):  # strip only ending chars.
                        if ignore_case and not regex:
                            # remove the last instance of frm_ from the string (case in-sensitive).
                            try:
                                s = re.sub(
                                    r"(.*)" + re.escape(frm_),
                                    r"\1",
                                    orig_str,
                                    flags=re.IGNORECASE,
                                )
                            except re.error:
                                s = orig_str  # Fallback if regex fails
                        else:
                            # remove last instance of frm_ from the string.
                            try:
                                s = "".join(orig_str.rsplit(frm_, 1))
                            except ValueError:
                                s = orig_str  # Fallback if empty separator

                    else:
                        if ignore_case and not regex:
                            # remove frm_ from the string (case in-sensitive).
                            try:
                                s = re.sub(
                                    re.escape(frm_), "", orig_str, flags=re.IGNORECASE
                                )
                            except re.error:
                                s = orig_str  # Fallback if regex fails
                        else:
                            s = orig_str.replace(
                                frm_, ""
                            )  # remove frm_ from the string.
                else:
                    s = orig_str  # No change if no filter pattern

            else:  # else; replace whole string.
                s = to_

            if return_orig_strings:
                result.append((orig_str, s))
            else:
                result.append(s)

        return result

    @staticmethod
    def format_suffix(
        string: str,
        suffix: str = "",
        strip: Union[str, List[str]] = "",
        strip_trailing_ints: bool = False,
        strip_trailing_alpha: bool = False,
    ) -> str:
        """Re-format the suffix for the given string.

        Parameters:
            string (str): The string to format.
            suffix (str): Append a new suffix to the given string.
            strip (str/list): Specific string(s) or regex pattern(s) to strip from the end of the given string.
            strip_trailing_ints (bool): Strip all trailing integers.
            strip_trailing_alpha (bool): Strip all upper-case letters preceded by a non-alphanumeric character.

        Returns:
            (str): The formatted string.
        """
        import re

        def is_regex(pattern: str) -> bool:
            try:
                re.compile(pattern)
                return True
            except re.error:
                return False

        # Always operate on the last pipe segment (Maya naming)
        s = string.split("|")[-1]

        if strip:
            strip_items = iter_utils.IterUtils.make_iterable(strip)
            for pattern in strip_items:
                if isinstance(pattern, str) and is_regex(pattern) and len(pattern) > 1:
                    # Only treat as regex if it is a pattern (not a simple suffix string)
                    s = re.sub(pattern, "", s)
                else:
                    # Standard: strip all occurrences of this suffix from the end
                    while s.endswith(pattern):
                        s = s[: -len(pattern)]

        # Strip trailing ints or uppercase alphas if requested
        while True:
            stripped = False
            if strip_trailing_ints and s and s[-1].isdigit():
                s = re.sub(r"\d+$", "", s)
                stripped = True
            if strip_trailing_alpha and s and s[-1].isupper():
                s = re.sub(r"[A-Z]+$", "", s)
                stripped = True
            if not stripped:
                break

        return s + suffix

    @staticmethod
    @core_utils.CoreUtils.listify(threading=True)
    def time_stamp(filepath, stamp="%m-%d-%Y  %H:%M"):
        """Attach or detach a modified timestamp and date to/from a given file path.

        Parameters:
            filepath (str): The full path to a file. ie. 'C:/Windows/Temp/__AUTO-SAVE__untitled.0001.mb'
            stamp (str): The time stamp format.

        Returns:
            str: Filepath with attached or detached timestamp, depending on whether it initially had a timestamp.
            ie. '16:46  11-09-2021  C:/Windows/Temp/__AUTO-SAVE__untitled.0001.mb' from 'C:/Windows/Temp/__AUTO-SAVE__untitled.0001.mb'
        """
        from datetime import datetime
        import os.path
        import re

        filepath = file_utils.FileUtils.format_path(filepath)

        # Check if the file path has a timestamp using regular expression
        match = re.match(r"\d{2}:\d{2}  \d{2}-\d{2}-\d{4}", filepath)
        if match:
            # If it does, return the file path without the timestamp
            return "".join(filepath.split()[2:])
        else:
            # If it doesn't, attach a timestamp
            try:
                return "{}  {}".format(
                    datetime.fromtimestamp(os.path.getmtime(filepath)).strftime(stamp),
                    filepath,
                )
            except (FileNotFoundError, OSError) as error:
                print(f"Error: {error}")
                return filepath


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------


# deprecated ---------------------



# !/usr/bin/python
# coding=utf-8
from typing import List, Optional, Tuple, Dict
import re


class FuzzyMatcher:
    """Fuzzy matching utilities for object names and hierarchical structures."""

    @staticmethod
    def get_base_name(name: str) -> str:
        """Remove trailing digits from name to get base name.

        Parameters:
            name: Input string that may have trailing digits

        Returns:
            String with trailing digits removed

        Examples:
            >>> FuzzyMatcher.get_base_name("object123")
            'object'
            >>> FuzzyMatcher.get_base_name("mesh_01")
            'mesh_'
        """
        return re.sub(r"\d+$", "", name)

    @staticmethod
    def find_best_match(
        target_name: str, available_names: List[str], score_threshold: float = 0.5
    ) -> Optional[Tuple[str, float]]:
        """Find best fuzzy match for target name from available candidates.

        Parameters:
            target_name: Name to find a match for
            available_names: List of candidate names to match against
            score_threshold: Minimum similarity score (0.0-1.0) to consider a match

        Returns:
            Tuple of (best_match, score) or None if no good match found

        Examples:
            >>> candidates = ["mesh_01", "mesh_02", "cube_01"]
            >>> FuzzyMatcher.find_best_match("mesh_03", candidates)
            ('mesh_01', 0.9)
        """
        # Try exact match first
        if target_name in available_names:
            return target_name, 1.0

        best_match = None
        best_score = 0.0

        for candidate in available_names:
            score = FuzzyMatcher._calculate_similarity(target_name, candidate)
            if score > best_score and score >= score_threshold:
                best_match = candidate
                best_score = score

        return (best_match, best_score) if best_match else None

    @staticmethod
    def find_all_matches(
        target_names: List[str],
        available_names: List[str],
        score_threshold: float = 0.5,
    ) -> Dict[str, Tuple[str, float]]:
        """Find fuzzy matches for multiple target names.

        Parameters:
            target_names: List of names to find matches for
            available_names: List of candidate names to match against
            score_threshold: Minimum similarity score to consider a match

        Returns:
            Dictionary mapping target_name -> (best_match, score)
        """
        matches = {}
        for target in target_names:
            match = FuzzyMatcher.find_best_match(
                target, available_names, score_threshold
            )
            if match:
                matches[target] = match
        return matches

    @staticmethod
    def find_trailing_digit_matches(
        missing_paths: List[str], extra_paths: List[str], path_separator: str = "|"
    ) -> Tuple[List[Dict[str, str]], List[str], List[str]]:
        """Find fuzzy matches specifically for trailing digit differences in hierarchical paths.

        This is useful for matching objects that have been renamed with different numbering
        but are otherwise identical (e.g., "mesh_01" vs "mesh_02").

        Parameters:
            missing_paths: List of paths that are missing
            extra_paths: List of paths that are extra/unexpected
            path_separator: Character used to separate hierarchy levels

        Returns:
            Tuple of:
            - fuzzy_matches: List of match dictionaries
            - paths_to_remove_from_missing: Paths that were matched and should be removed
            - paths_to_remove_from_extra: Paths that were matched and should be removed
        """
        missing_to_check = missing_paths.copy()
        extra_to_check = extra_paths.copy()

        fuzzy_matches = []
        paths_to_remove_from_missing = []
        paths_to_remove_from_extra = []

        for missing_path in missing_to_check:
            missing_node_name = missing_path.split(path_separator)[-1]
            missing_base_name = FuzzyMatcher.get_base_name(missing_node_name)

            # Skip if no trailing digits
            if missing_base_name == missing_node_name:
                continue

            for extra_path in extra_to_check:
                extra_node_name = extra_path.split(path_separator)[-1]
                extra_base_name = FuzzyMatcher.get_base_name(extra_node_name)

                if (
                    missing_base_name == extra_base_name
                    and missing_node_name != extra_node_name
                ):

                    # Check if they have the same parent
                    missing_parent = path_separator.join(
                        missing_path.split(path_separator)[:-1]
                    )
                    extra_parent = path_separator.join(
                        extra_path.split(path_separator)[:-1]
                    )

                    if missing_parent == extra_parent:
                        fuzzy_match = {
                            "current_name": extra_node_name,
                            "current_path": extra_path,
                            "target_name": missing_node_name,
                            "target_path": missing_path,
                            "parent_path": missing_parent,
                        }
                        fuzzy_matches.append(fuzzy_match)
                        paths_to_remove_from_missing.append(missing_path)
                        paths_to_remove_from_extra.append(extra_path)
                        break

        return fuzzy_matches, paths_to_remove_from_missing, paths_to_remove_from_extra

    @staticmethod
    def _calculate_similarity(name1: str, name2: str) -> float:
        """Calculate similarity score between two names using multiple strategies.

        Parameters:
            name1: First name to compare
            name2: Second name to compare

        Returns:
            Similarity score from 0.0 (no similarity) to 1.0 (identical)
        """
        # Strategy 1: Base name matching (remove trailing digits)
        base1 = FuzzyMatcher.get_base_name(name1)
        base2 = FuzzyMatcher.get_base_name(name2)

        if base1 == base2 and base1:
            # Same base name - check if at least one has trailing digits
            if base1 != name1 or base1 != name2:
                # At least one has trailing digits and same base - high score
                return 0.9
            # Both names are identical to base (no trailing digits) - perfect match
            elif name1 == name2:
                return 1.0

        # Strategy 2: Substring matching
        if name1 in name2 or name2 in name1:
            return min(len(name1), len(name2)) / max(len(name1), len(name2))

        # Strategy 3: Common prefix length
        common_prefix = 0
        for a, b in zip(name1, name2):
            if a == b:
                common_prefix += 1
            else:
                break

        if common_prefix > 5:  # Arbitrary threshold for meaningful prefix
            return common_prefix / max(len(name1), len(name2))

        return 0.0

    @staticmethod
    def calculate_levenshtein_distance(s1: str, s2: str) -> int:
        """Calculate Levenshtein (edit) distance between two strings.

        Parameters:
            s1: First string
            s2: Second string

        Returns:
            Number of single-character edits needed to change s1 into s2
        """
        if len(s1) < len(s2):
            return FuzzyMatcher.calculate_levenshtein_distance(s2, s1)

        if len(s2) == 0:
            return len(s1)

        previous_row = list(range(len(s2) + 1))
        for i, c1 in enumerate(s1):
            current_row = [i + 1]
            for j, c2 in enumerate(s2):
                insertions = previous_row[j + 1] + 1
                deletions = current_row[j] + 1
                substitutions = previous_row[j] + (c1 != c2)
                current_row.append(min(insertions, deletions, substitutions))
            previous_row = current_row

        return previous_row[-1]

    @staticmethod
    def similarity_from_distance(s1: str, s2: str) -> float:
        """Calculate similarity score from Levenshtein distance.

        Parameters:
            s1: First string
            s2: Second string

        Returns:
            Similarity score from 0.0 to 1.0
        """
        distance = FuzzyMatcher.calculate_levenshtein_distance(s1, s2)
        max_len = max(len(s1), len(s2))
        return 1.0 - (distance / max_len) if max_len > 0 else 1.0


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    # Example usage and tests
    matcher = FuzzyMatcher()

    # Test basic functionality
    candidates = ["mesh_01", "mesh_02", "cube_01", "sphere_03"]

    print("=== Fuzzy Matching Examples ===")
    print(f"Candidates: {candidates}")

    test_names = ["mesh_03", "cube_02", "plane_01"]
    for name in test_names:
        match = matcher.find_best_match(name, candidates)
        if match:
            print(f"'{name}' -> '{match[0]}' (score: {match[1]:.2f})")
        else:
            print(f"'{name}' -> No good match found")

    # Test hierarchical path matching
    print("\n=== Hierarchical Path Matching ===")
    missing = ["group1|mesh_01", "group2|cube_02"]
    extra = ["group1|mesh_03", "group2|cube_05", "group3|sphere_01"]

    matches, missing_remove, extra_remove = matcher.find_trailing_digit_matches(
        missing, extra
    )

    for match in matches:
        print(f"Fuzzy match: {match['current_name']} -> {match['target_name']}")

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------
# This module provides general-purpose fuzzy string matching utilities that can be used
# across different domains, not just Maya. It's particularly useful for:
#
# - Object name matching with numerical variations
# - Hierarchical path matching
# - String similarity calculations
# - Levenshtein distance calculations
#
# The module is designed to be imported and used by domain-specific tools like Maya
# hierarchy managers, but the core algorithms are generic.
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
from collections import namedtuple
from typing import List, Dict, Any, Optional, Union, Iterator, Callable
from pythontk.core_utils import LoggingMixin


class NamedTupleContainer(LoggingMixin):
    """A generic container class for managing collections of named tuples.

    The class provides methods to query, modify, extend, and remove elements within the container.
    It supports dynamic field access, duplicate handling, and flexible extension mechanisms.

    Attributes:
        named_tuples (List): The list of named tuples stored in this container.
        metadata (Dict[str, Any]): A dictionary containing additional information like "fields"
            which are the named tuple fields, and an "allow_duplicates" flag.
        fields (List[str]): List of named tuple fields derived from metadata.
        _tuple_class: The dynamically generated named tuple class based on `fields`.

    Methods:
        extend: Add new named tuples to the container while handling duplicates.
        get: Query the named tuples based on certain conditions and optionally retrieve a specific field's value.
        modify: Modify a named tuple at a specific index.
        remove: Remove a named tuple at a specific index.

    Examples:
        Basic usage:
        ```python
        container = NamedTupleContainer(
            named_tuples=[],
            fields=["name", "age", "email"],
            metadata={"allow_duplicates": False}
        )

        # Add some data
        Person = namedtuple("Person", ["name", "age", "email"])
        people = [
            Person("Alice", 30, "alice@example.com"),
            Person("Bob", 25, "bob@example.com")
        ]
        container.extend(people)

        # Query data
        adults = container.get(age=30)
        names = container.name  # Get all names via dynamic attribute access
        ```

        With custom extender function:
        ```python
        def file_extender(container, objects, **metadata):
            # Custom logic to process objects and return list of tuples
            return [(obj.name, obj.path) for obj in objects]

        container = NamedTupleContainer(
            extender_func=file_extender,
            fields=["filename", "filepath"]
        )
        container.extend(file_objects)
        ```

    Notes:
        This class utilizes internal logging. The logging level can be set during instantiation.
        The class defines custom `__iter__` and `__repr__` methods to iterate over the named tuples and represent the object.
        Uses the `namedtuple` class from Python's standard library to dynamically create tuple classes.
    """

    def __init__(
        self,
        named_tuples: Optional[List[namedtuple]] = None,
        fields: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        extender_func: Optional[Callable] = None,
        tuple_class_name: str = "TupleClass",
        log_level: str = "WARNING",
    ) -> None:
        """
        Creates a container for named tuples, providing dynamic attribute access and query capabilities.

        Args:
            named_tuples: A list of named tuples to initialize the container with.
            fields: List of field names for the named tuples.
            metadata: Metadata related to the container, including field names and other settings.
            extender_func: Optional function to handle extending the container with new objects.
                Should have signature: func(container, objects, **metadata) -> List[tuple]
            tuple_class_name: Name for the dynamically created tuple class.
            log_level: Logging level. Defaults to "WARNING".
        """
        self.logger.setLevel(log_level)

        self.named_tuples = named_tuples or []
        self.metadata = metadata or {}

        # Fields can come from parameter, metadata, or inferred from existing tuples
        if fields:
            self.fields = fields
        elif "fields" in self.metadata:
            from pythontk import make_iterable

            self.fields = make_iterable(self.metadata["fields"])
        elif self.named_tuples:
            # Infer from first tuple
            self.fields = list(self.named_tuples[0]._fields)
        else:
            self.fields = []

        self.extender_func = extender_func
        self._tuple_class = (
            namedtuple(tuple_class_name, self.fields) if self.fields else None
        )

    def __iter__(self) -> Iterator[namedtuple]:
        """
        Allows iteration over the named tuples in the container.

        Returns:
            An iterator over the named tuples.
        """
        return iter(self.named_tuples)

    def __repr__(self) -> str:
        """
        Returns the string representation of the named tuples.

        Returns:
            The string representation of the named tuples.
        """
        return f"<NamedTupleContainer({len(self.named_tuples)} items, fields={self.fields})>"

    def __len__(self) -> int:
        """
        Returns the number of named tuples in the container.

        Returns:
            The count of named tuples.
        """
        return len(self.named_tuples)

    def __getattr__(self, name: str) -> Any:
        """
        Dynamic attribute access for field names.

        Args:
            name: The attribute name.

        Returns:
            A list of values for the specified field name if found in fields.

        Raises:
            AttributeError: If the attribute is not found in fields.
        """
        if name in self.fields:
            return [getattr(nt, name) for nt in self.named_tuples]
        else:
            raise AttributeError(
                f"'NamedTupleContainer' object has no attribute '{name}'"
            )

    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[namedtuple, List[namedtuple]]:
        """
        Get named tuple(s) by index or slice.

        Args:
            index: Integer index or slice object.

        Returns:
            Single named tuple or list of named tuples.
        """
        return self.named_tuples[index]

    def __setitem__(self, index: int, value: namedtuple) -> None:
        """
        Set a named tuple at a specific index.

        Args:
            index: The index to set.
            value: The named tuple to set.
        """
        self.named_tuples[index] = value

    def __delitem__(self, index: int) -> None:
        """
        Delete a named tuple at a specific index.

        Args:
            index: The index to delete.
        """
        del self.named_tuples[index]

    @staticmethod
    def _handle_duplicates(
        existing: List[namedtuple],
        new: List[namedtuple],
        allow_duplicates: bool,
        signature_func: Optional[Callable] = None,
    ) -> List[namedtuple]:
        """
        Handles duplicates based on the allow_duplicates flag.

        Args:
            existing: List of existing named tuples.
            new: List of new named tuples to add.
            allow_duplicates: Flag to allow or disallow duplicates.
            signature_func: Optional function to create signatures for comparison.

        Returns:
            Combined list of named tuples with or without duplicates based on the flag.
        """
        if allow_duplicates:
            return existing + new

        def default_signature(nt):
            """Default signature creation that handles class objects intelligently"""
            if hasattr(nt, "_fields"):
                # Named tuple
                sig = []
                for field_name in nt._fields:
                    value = getattr(nt, field_name)
                    # For class objects, use the class name and module instead of the object itself
                    if hasattr(value, "__name__") and hasattr(value, "__module__"):
                        sig.append((field_name, value.__name__, value.__module__))
                    else:
                        sig.append((field_name, value))
                return tuple(sig)
            else:
                # Regular tuple - just return as is
                return nt

        signature_func = signature_func or default_signature

        # Create signatures for existing tuples
        existing_signatures = {signature_func(nt) for nt in existing}

        # Only add new tuples that don't match existing signatures
        unique_new = []
        for nt in new:
            if signature_func(nt) not in existing_signatures:
                unique_new.append(nt)
                existing_signatures.add(signature_func(nt))

        return existing + unique_new

    def extend(
        self, objects: Union[List[namedtuple], List[tuple], Any], **metadata
    ) -> None:
        """
        Extend the container with new objects while handling duplicates properly.

        Args:
            objects: Objects to add. Can be:
                - List of named tuples (added directly)
                - List of tuples (converted to named tuples)
                - Other objects (processed by extender_func if provided)
            **metadata: Additional metadata to merge with existing metadata.

        Raises:
            ValueError: If no way to process the objects is available.
            Exception: For other errors during extension.
        """
        try:
            # Merge metadata
            merged_metadata = {**self.metadata, **metadata}
            allow_duplicates = merged_metadata.get("allow_duplicates", False)
            signature_func = merged_metadata.get("signature_func")

            new_named_tuples = []

            # Handle different types of input objects
            if isinstance(objects, list) and objects:
                first_obj = objects[0]

                # Case 1: Already named tuples
                if hasattr(first_obj, "_fields"):
                    new_named_tuples = objects

                # Case 2: Regular tuples - convert to named tuples
                elif isinstance(first_obj, tuple):
                    if self._tuple_class:
                        new_named_tuples = [self._tuple_class(*obj) for obj in objects]
                    else:
                        raise ValueError("No tuple class available to convert tuples")

                # Case 3: Other objects - use extender function
                else:
                    if self.extender_func:
                        tuple_data = self.extender_func(
                            self, objects, **merged_metadata
                        )
                        if self._tuple_class:
                            new_named_tuples = [
                                self._tuple_class(*data) for data in tuple_data
                            ]
                        else:
                            raise ValueError(
                                "No tuple class available for extender function results"
                            )
                    else:
                        raise ValueError(
                            "No extender function provided for processing objects"
                        )

            # Handle single object
            elif objects is not None:
                return self.extend([objects], **metadata)

            # Apply duplicate handling and extend the container
            if new_named_tuples:
                # Ensure we use the same tuple class as existing tuples
                if self.named_tuples and new_named_tuples:
                    existing_class = type(self.named_tuples[0])
                    new_class = type(new_named_tuples[0])

                    # Convert to same class if different
                    if existing_class != new_class:
                        new_named_tuples = [
                            existing_class(*nt) for nt in new_named_tuples
                        ]

                self.named_tuples = self._handle_duplicates(
                    self.named_tuples,
                    new_named_tuples,
                    allow_duplicates,
                    signature_func,
                )

        except Exception as e:
            self.logger.error(
                f"An error occurred while extending the container: {str(e)}"
            )
            raise

    def get(
        self, return_field: Optional[str] = None, **conditions
    ) -> Union[List[Any], Any, None]:
        """
        Query the named tuples based on specified conditions.

        Args:
            return_field: The name of the field to return. If None, returns the entire named tuple.
            **conditions: Key-value pairs representing the query conditions.

        Returns:
            A list of matching named tuples or specified field values.
            If conditions and return_field are specified, returns the first matching value or None if not found.
        """
        results = []
        for named_tuple in self.named_tuples:
            if all(
                hasattr(named_tuple, field) and getattr(named_tuple, field) == value
                for field, value in conditions.items()
            ):
                result = (
                    getattr(named_tuple, return_field) if return_field else named_tuple
                )
                # If conditions and return_field are specified, return the first match
                if conditions and return_field:
                    return result
                results.append(result)
        return results

    def filter(self, predicate: Callable[[namedtuple], bool]) -> "NamedTupleContainer":
        """
        Filter the container based on a predicate function.

        Args:
            predicate: Function that takes a named tuple and returns True/False.

        Returns:
            A new NamedTupleContainer with filtered results.
        """
        filtered_tuples = [nt for nt in self.named_tuples if predicate(nt)]
        return NamedTupleContainer(
            named_tuples=filtered_tuples,
            fields=self.fields,
            metadata=self.metadata.copy(),
            extender_func=self.extender_func,
        )

    def map(self, func: Callable[[namedtuple], namedtuple]) -> "NamedTupleContainer":
        """
        Apply a function to all named tuples in the container.

        Args:
            func: Function that takes a named tuple and returns a modified named tuple.

        Returns:
            A new NamedTupleContainer with transformed results.
        """
        mapped_tuples = [func(nt) for nt in self.named_tuples]
        return NamedTupleContainer(
            named_tuples=mapped_tuples,
            fields=self.fields,
            metadata=self.metadata.copy(),
            extender_func=self.extender_func,
        )

    def modify(self, index: int, **kwargs) -> namedtuple:
        """
        Modify a named tuple at a specific index within the container.

        Args:
            index: The index of the named tuple within the container to modify.
            **kwargs: Key-value pairs representing the fields to update and their new values.

        Returns:
            The updated named tuple.

        Raises:
            IndexError: If the index is out of range.
            AttributeError: If trying to modify a field that doesn't exist.
        """
        if not 0 <= index < len(self.named_tuples):
            raise IndexError(
                f"Index {index} is out of range for container with {len(self.named_tuples)} items"
            )

        named_tuple = self.named_tuples[index]

        # Validate that all kwargs fields exist in the named tuple
        for field in kwargs:
            if not hasattr(named_tuple, field):
                raise AttributeError(f"Field '{field}' does not exist in named tuple")

        new_tuple = named_tuple._replace(**kwargs)
        self.named_tuples[index] = new_tuple
        return new_tuple

    def remove(self, index: int) -> namedtuple:
        """
        Remove a named tuple at a specific index within the container.

        Args:
            index: The index of the named tuple within the container to remove.

        Returns:
            The removed named tuple.

        Raises:
            IndexError: If the index is out of range.
        """
        if not 0 <= index < len(self.named_tuples):
            raise IndexError(
                f"Index {index} is out of range for container with {len(self.named_tuples)} items"
            )

        return self.named_tuples.pop(index)

    def clear(self) -> None:
        """Clear all named tuples from the container."""
        self.named_tuples.clear()

    def to_dict_list(self) -> List[Dict[str, Any]]:
        """
        Convert all named tuples to a list of dictionaries.

        Returns:
            List of dictionaries representing the named tuples.
        """
        return [nt._asdict() for nt in self.named_tuples]

    def to_csv(self, filename: str, **kwargs) -> None:
        """
        Export the container to a CSV file.

        Args:
            filename: Path to the CSV file to write.
            **kwargs: Additional arguments passed to csv.writer.
        """
        import csv

        with open(filename, "w", newline="", **kwargs) as csvfile:
            if self.named_tuples:
                writer = csv.DictWriter(csvfile, fieldnames=self.fields)
                writer.writeheader()
                for nt in self.named_tuples:
                    writer.writerow(nt._asdict())

    @classmethod
    def from_csv(
        cls, filename: str, tuple_class_name: str = "CsvRow", **kwargs
    ) -> "NamedTupleContainer":
        """
        Create a NamedTupleContainer from a CSV file.

        Args:
            filename: Path to the CSV file to read.
            tuple_class_name: Name for the created tuple class.
            **kwargs: Additional arguments passed to csv.reader.

        Returns:
            A new NamedTupleContainer with data from the CSV.
        """
        import csv

        with open(filename, "r", **kwargs) as csvfile:
            reader = csv.DictReader(csvfile)
            fields = reader.fieldnames

            if not fields:
                return cls(fields=[], tuple_class_name=tuple_class_name)

            TupleClass = namedtuple(tuple_class_name, fields)
            named_tuples = [TupleClass(**row) for row in reader]

            return cls(
                named_tuples=named_tuples,
                fields=fields,
                tuple_class_name=tuple_class_name,
            )


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    # Example usage demonstrations

    # Basic usage with direct tuple creation
    print("=== Basic Usage ===")
    Person = namedtuple("Person", ["name", "age", "city"])
    people = [
        Person("Alice", 30, "New York"),
        Person("Bob", 25, "London"),
        Person("Charlie", 35, "Tokyo"),
    ]

    container = NamedTupleContainer(named_tuples=people)
    print(f"Container: {container}")
    print(f"Names: {container.name}")
    print(f"Adults over 30: {container.get(age=35)}")

    # Usage with extender function
    print("\n=== Custom Extender Function ===")

    def file_data_extender(container, objects, **metadata):
        """Example extender that processes file-like objects"""
        return [(obj, f"/path/to/{obj}", "file") for obj in objects]

    file_container = NamedTupleContainer(
        fields=["name", "path", "type"],
        extender_func=file_data_extender,
        tuple_class_name="FileInfo",
    )

    file_container.extend(["file1.txt", "file2.py", "file3.md"])
    print(f"File container: {file_container}")
    print(f"File paths: {file_container.path}")


# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------
#
# This standalone NamedTupleContainer provides:
# 1. Generic tuple management capabilities
# 2. Flexible extension mechanisms via custom extender functions
# 3. Advanced querying and filtering
# 4. Export/import capabilities (CSV)
# 5. Functional programming style methods (map, filter)
# 6. Full compatibility with the original FileManager use case
#
# The key insight is that the container itself is domain-agnostic -
# it's the extender function that provides domain-specific logic.
# This makes it reusable across many different use cases beyond file management.
#



# !/usr/bin/python
# coding=utf-8
from typing import Dict, Optional, Any


class ClassProperty:
    """A descriptor for class-level properties (replaces @classmethod @property).

    This decorator allows you to define a property that can be accessed directly on the class,
    rather than on an instance of the class. This is useful for defining properties that are
    related to the class itself rather than to a specific instance.

    Example:
        class MyClass:
            _value = 42

            @ClassProperty
            def value(cls):
                return cls._value

        print(MyClass.value)  # Output: 42
        MyClass._value = 100
    """

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


# --------------------------------------------------------------------------------------------


if __name__ == "__main__":
    import unittest

    class TestClassProperty(unittest.TestCase):
        class MyClass:
            _value = 42

            @ClassProperty
            def value(cls):
                return cls._value

        def test_class_property(self):
            self.assertEqual(self.MyClass.value, 42)

        def test_class_property_setter(self):
            self.MyClass._value = 100
            self.assertEqual(self.MyClass.value, 100)

    unittest.main(exit=False)


# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
from typing import List, Dict, Any, Optional, Callable, Tuple
from .hierarchy_indexer import HierarchyIndexer


class HierarchyMatching:
    """Generic matching strategies for hierarchical data."""

    @staticmethod
    def _clean_namespace(name: str, separator: str = ":") -> str:
        """Remove namespace prefix from a name."""
        return name.split(separator)[-1] if separator in name else name

    @staticmethod
    def _get_path_tail(path: str, num_components: int = 1, separator: str = "|") -> str:
        """Get the last N components of a hierarchy path."""
        components = path.split(separator) if path else []
        tail_components = components[-num_components:] if components else []
        return separator.join(tail_components)

    @staticmethod
    def _path_ends_with(path: str, suffix: str, separator: str = "|") -> bool:
        """Check if a hierarchical path ends with the given suffix."""
        if not suffix:
            return True
        path_components = path.split(separator)
        suffix_components = suffix.split(separator)

        if len(suffix_components) > len(path_components):
            return False

        return path_components[-len(suffix_components) :] == suffix_components

    @staticmethod
    def exact_path_match(
        source_items: List[Any],
        target_items: List[Any],
        get_path_func: Callable[[Any], str],
        path_separator: str = "|",
        clean_namespaces: bool = True,
        namespace_separator: str = ":",
    ) -> Dict[Any, List[Any]]:
        """Find exact path matches between source and target items.

        Args:
            source_items: Items to find matches for
            target_items: Items to search within
            get_path_func: Function to extract path from an item
            path_separator: Character separating path components
            clean_namespaces: Whether to remove namespace prefixes
            namespace_separator: Character separating namespace from name

        Returns:
            Dictionary mapping source items to lists of matching target items
        """
        # Build index of target items
        target_index = HierarchyIndexer.build_path_index(
            target_items,
            get_path_func,
            path_separator,
            clean_namespaces,
            namespace_separator,
        )

        matches = {}
        for source_item in source_items:
            source_path = get_path_func(source_item)
            if source_path:
                target_matches = HierarchyIndexer.find_by_path(
                    target_index,
                    source_path,
                    clean_namespaces,
                    path_separator,
                    namespace_separator,
                )
                if target_matches:
                    matches[source_item] = target_matches

        return matches

    @staticmethod
    def tail_path_match(
        source_items: List[Any],
        target_items: List[Any],
        get_path_func: Callable[[Any], str],
        num_components: int = 2,
        path_separator: str = "|",
        clean_namespaces: bool = True,
        namespace_separator: str = ":",
    ) -> Dict[Any, List[Any]]:
        """Find matches by comparing tail portions of paths.

        Args:
            source_items: Items to find matches for
            target_items: Items to search within
            get_path_func: Function to extract path from an item
            num_components: Number of tail components to compare
            path_separator: Character separating path components
            clean_namespaces: Whether to remove namespace prefixes
            namespace_separator: Character separating namespace from name

        Returns:
            Dictionary mapping source items to lists of matching target items
        """
        # Build index of target items
        target_index = HierarchyIndexer.build_path_index(
            target_items,
            get_path_func,
            path_separator,
            clean_namespaces,
            namespace_separator,
        )

        matches = {}
        for source_item in source_items:
            source_path = get_path_func(source_item)
            if source_path:
                normalized_path = HierarchyIndexer._normalize_path(
                    source_path, clean_namespaces, path_separator, namespace_separator
                )
                source_tail = HierarchyMatching._get_path_tail(
                    normalized_path, num_components, path_separator
                )

                target_matches = HierarchyIndexer.find_by_tail_path(
                    target_index, source_tail, num_components, path_separator
                )
                if target_matches:
                    matches[source_item] = target_matches

        return matches

    @staticmethod
    def fuzzy_name_match(
        source_items: List[Any],
        target_items: List[Any],
        get_name_func: Callable[[Any], str],
        similarity_threshold: float = 0.8,
    ) -> Dict[Any, List[Any]]:
        """Find matches using fuzzy string matching on names.

        Args:
            source_items: Items to find matches for
            target_items: Items to search within
            get_name_func: Function to extract name from an item
            similarity_threshold: Minimum similarity score (0.0 to 1.0)

        Returns:
            Dictionary mapping source items to lists of matching target items
        """
        try:
            from difflib import SequenceMatcher
        except ImportError:
            # Fallback to simple exact matching if difflib not available
            return HierarchyMatching._exact_name_match(
                source_items, target_items, get_name_func
            )

        matches = {}

        for source_item in source_items:
            source_name = get_name_func(source_item)
            if not source_name:
                continue

            source_clean = HierarchyMatching._clean_namespace(source_name)
            fuzzy_matches = []

            for target_item in target_items:
                target_name = get_name_func(target_item)
                if not target_name:
                    continue

                target_clean = HierarchyMatching._clean_namespace(target_name)

                # Calculate similarity
                similarity = SequenceMatcher(None, source_clean, target_clean).ratio()
                if similarity >= similarity_threshold:
                    fuzzy_matches.append(target_item)

            if fuzzy_matches:
                matches[source_item] = fuzzy_matches

        return matches

    @staticmethod
    def _exact_name_match(
        source_items: List[Any],
        target_items: List[Any],
        get_name_func: Callable[[Any], str],
    ) -> Dict[Any, List[Any]]:
        """Fallback exact name matching when fuzzy matching is unavailable."""
        # Build index of target items by clean name
        target_index = {}
        for target_item in target_items:
            target_name = get_name_func(target_item)
            if target_name:
                clean_name = HierarchyMatching._clean_namespace(target_name)
                if clean_name not in target_index:
                    target_index[clean_name] = []
                target_index[clean_name].append(target_item)

        matches = {}
        for source_item in source_items:
            source_name = get_name_func(source_item)
            if source_name:
                clean_name = HierarchyMatching._clean_namespace(source_name)
                if clean_name in target_index:
                    matches[source_item] = target_index[clean_name]

        return matches

    @staticmethod
    def multi_strategy_match(
        source_items: List[Any],
        target_items: List[Any],
        get_path_func: Callable[[Any], str],
        get_name_func: Optional[Callable[[Any], str]] = None,
        strategies: List[str] = None,
        path_separator: str = "|",
        clean_namespaces: bool = True,
        namespace_separator: str = ":",
        fuzzy_threshold: float = 0.8,
    ) -> Dict[Any, List[Any]]:
        """Apply multiple matching strategies in order of preference.

        Args:
            source_items: Items to find matches for
            target_items: Items to search within
            get_path_func: Function to extract path from an item
            get_name_func: Function to extract name from an item (for fuzzy matching)
            strategies: List of strategy names to try in order
            path_separator: Character separating path components
            clean_namespaces: Whether to remove namespace prefixes
            namespace_separator: Character separating namespace from name
            fuzzy_threshold: Minimum similarity for fuzzy matching

        Returns:
            Dictionary mapping source items to lists of matching target items
        """
        if strategies is None:
            strategies = ["exact_path", "tail_path", "fuzzy_name"]

        all_matches = {}
        unmatched_items = list(source_items)

        for strategy in strategies:
            if not unmatched_items:
                break

            if strategy == "exact_path":
                matches = HierarchyMatching.exact_path_match(
                    unmatched_items,
                    target_items,
                    get_path_func,
                    path_separator,
                    clean_namespaces,
                    namespace_separator,
                )
            elif strategy == "tail_path":
                matches = HierarchyMatching.tail_path_match(
                    unmatched_items,
                    target_items,
                    get_path_func,
                    2,
                    path_separator,
                    clean_namespaces,
                    namespace_separator,
                )
            elif strategy == "fuzzy_name" and get_name_func:
                matches = HierarchyMatching.fuzzy_name_match(
                    unmatched_items, target_items, get_name_func, fuzzy_threshold
                )
            else:
                continue

            # Add matches and remove matched items from unmatched list
            for source_item, target_matches in matches.items():
                all_matches[source_item] = target_matches
                if source_item in unmatched_items:
                    unmatched_items.remove(source_item)

        return all_matches


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    ...

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import sys
import re
import json
import subprocess
from pythontk.core_utils import help_mixin


class PkgManagerHelperMixin:
    """A mixin class to provide package management capabilities using pip.

    This mixin class offers methods to interact with the Python Package Index (PyPI)
    using pip. It allows for the installation, uninstallation, listing, and updating
    of Python packages, as well as retrieving package details and versions.

    Methods:
        install: Install a specified Python package.
        uninstall: Uninstall a specified Python package.
        list_packages: List all installed Python packages.
        package_details: Retrieve detailed information about a specified package.
        update: Update a specified Python package to the latest version.
        package_version: Get the installed version of a specified package.
        list_outdated_packages: List all outdated Python packages in the environment.
        is_outdated: Check if a specified package is outdated.

    The class is intended to be used as a mixin, providing package management functionalities
    to other classes that deal with Python environments and package management.
    """

    def install(self, package_name):
        """Install a package.

        Example:
            pkg_mgr.install("numpy")
        """
        self.pip(f"install {package_name}")

    def uninstall(self, package_name):
        """Uninstall a package.

        Example:
            pkg_mgr.uninstall("numpy")
        """
        self.pip(f"uninstall {package_name} -y")

    def list_packages(self):
        """List installed packages.

        Example:
            installed_packages = pkg_mgr.list_packages()
        """
        return self.pip("list")

    def package_details(self, package_name):
        """Show details of a specific package.

        Example:
            details = pkg_mgr.package_details("numpy")
        """
        return self.pip(f"show {package_name}")

    def update(self, package_name):
        """Update a package to the latest version.

        Example:
            pkg_mgr.update("numpy")
        """
        self.pip(f"install --upgrade {package_name}")

    def installed_version(self, package_name):
        """Get the installed version of a package."""
        package_info = self.package_details(package_name)
        return package_info.get("version", "Not installed")

    def latest_version(self, package_name):
        """Get the latest version of a package from PyPI using the standard library."""
        import json
        from urllib.request import Request, urlopen
        from urllib.error import URLError, HTTPError

        url = f"https://pypi.org/pypi/{package_name}/json"
        request = Request(url)

        try:
            with urlopen(request) as response:
                data = json.loads(response.read().decode())
                return data["info"]["version"]  # Return the latest version
        except HTTPError as e:
            raise RuntimeError(
                f"HTTP error occurred while fetching the latest version of {package_name}: {e.code} {e.reason}"
            )
        except URLError as e:
            raise RuntimeError(
                f"URL error occurred while fetching the latest version of {package_name}: {e.reason}"
            )
        except json.JSONDecodeError as e:
            raise RuntimeError(
                f"Failed to parse JSON response for {package_name}: {e.msg}"
            )

    def list_outdated_packages(self):
        """List all outdated packages.

        Example:
            outdated_packages = pkg_mgr.list_outdated_packages()
        """
        outdated_packages = self.pip("list --outdated --format=json")
        if isinstance(outdated_packages, str):
            try:
                return json.loads(outdated_packages)
            except json.JSONDecodeError:
                raise ValueError("Failed to decode JSON from outdated packages list")
        elif isinstance(outdated_packages, dict):
            # If the output is already a dictionary, return it as is or extract the relevant data
            return outdated_packages
        else:
            raise TypeError("Unexpected data type for outdated packages list")

    def is_outdated(self, package_name: str) -> bool:
        """Efficiently check if a specific package is outdated."""
        return self.installed_version(package_name) != self.latest_version(package_name)


class PkgVersionCheck:
    """A class to check package versions and detect available updates.

    This class provides functionality to check if a package is up-to-date by comparing
    the installed version against the latest version available on PyPI.

    Attributes:
        _package_name: The name of the package to check.
        _python_path: Path to the Python interpreter to use for checking.
        _installed_ver: Cached installed version of the package.
        _latest_ver: Cached latest version of the package from PyPI.
    """

    _package_name: str = None
    _python_path: str = None
    _installed_ver: str = ""
    _latest_ver: str = ""

    def __init__(self, package_name=None, python_path=None):
        """Initialize the PkgVersionCheck.

        Parameters:
            package_name (str, optional): The name of the package to check.
            python_path (str, optional): Path to the Python interpreter to use.
                Defaults to the current Python interpreter.
        """
        if package_name:
            self._package_name = package_name
        if python_path:
            self._python_path = python_path
        else:
            self._python_path = sys.executable

    def start_version_check(self, package_name=None, python_path=None) -> None:
        """Start a version check in a background thread.

        Parameters:
            package_name (str, optional): The name of the package to check.
                If not provided, uses the value set in __init__ or the class default.
            python_path (str, optional): Path to the Python interpreter to use.
                If not provided, uses the value set in __init__ or the class default.
        """
        if package_name:
            self._package_name = package_name
        if python_path:
            self._python_path = python_path

        # Make sure we have a package name
        if not self._package_name:
            raise ValueError("Package name must be provided")

        # Start a background thread to check versions
        import threading

        threading.Thread(target=self.check_version, daemon=True).start()

    @property
    def new_version_available(self) -> bool:
        """Check if a new version of the package is available.

        Returns:
            bool: True if a newer version is available, False otherwise.
        """
        try:
            return (
                self.installed_ver != self.latest_ver
                and self.installed_ver
                and self.latest_ver
            )
        except AttributeError:
            return False

    @property
    def installed_ver(self) -> str:
        """Get the installed version of the package.

        Returns:
            str: The installed version, or an empty string if not cached.
        """
        return getattr(self, "_installed_ver", "")

    @property
    def latest_ver(self) -> str:
        """Get the latest version of the package from PyPI.

        Returns:
            str: The latest version, or an empty string if not cached.
        """
        return getattr(self, "_latest_ver", "")

    def check_version(self, package_name=None, python_path=None) -> None:
        """Check the installed and latest versions of the package.

        This method updates the _installed_ver and _latest_ver attributes.

        Parameters:
            package_name (str, optional): The name of the package to check.
                If not provided, uses the value set in __init__ or the class default.
            python_path (str, optional): Path to the Python interpreter to use.
                If not provided, uses the value set in __init__ or the class default.
        """
        if package_name:
            self._package_name = package_name
        if python_path:
            self._python_path = python_path

        # Make sure we have a package name
        if not self._package_name:
            raise ValueError("Package name must be provided")

        # Create a package manager for the specified Python interpreter
        pkg_mgr = PkgManager(python_path=self._python_path)

        # Get the installed and latest versions
        self._installed_ver = pkg_mgr.installed_version(self._package_name)
        self._latest_ver = pkg_mgr.latest_version(self._package_name)


class PkgVersionUtils:
    """Utilities for managing package version information in files.

    This class provides static methods for updating version numbers in source files
    and managing package version specifications in requirements files.
    """

    @staticmethod
    def update_version(
        filepath: str,
        change: str = "increment",
        version_part: str = "patch",
        max_version_parts: tuple = (99, 99),
        version_regex: str = r"__version__\s*=\s*['\"](\d+)\.(\d+)\.(\d+)['\"]",
    ) -> str:
        """Update the version number in a text file.

        This function updates the version number in a text file depending on its state.
        The version number is represented as a string in the format 'x.y.z', where x, y,
        and z are integers and it matches the provided regex pattern.

        Args:
            filepath: The path to the text file containing the version number.
            change: The type of change, either 'increment' or 'decrement'. Defaults to 'increment'.
            version_part: The part of the version to update ('major', 'minor', or 'patch').
            max_version_parts: Maximum values for the minor and patch version parts.
            version_regex: Regex pattern for finding the version string.

        Returns:
            str: The new version number or empty string if not found.
        """
        import re
        import os
        from pythontk.file_utils import FileUtils

        lines = FileUtils.get_file_contents(filepath, as_list=True)

        version_pattern = re.compile(version_regex)
        max_minor, max_patch = max_version_parts

        version = ""
        for i, line in enumerate(lines):
            match = version_pattern.search(line)
            if match:
                major, minor, patch = map(int, match.groups())

                if version_part == "patch":
                    if change == "increment":
                        patch = (patch + 1) % (max_patch + 1)
                        if patch == 0:
                            minor = (minor + 1) % (max_minor + 1)
                            major += minor == 0
                    elif change == "decrement":
                        if patch == 0:
                            patch = max_patch
                            minor = (
                                (minor - 1) % (max_minor + 1)
                                if minor > 0
                                else max_minor
                            )
                            major -= minor == max_minor
                        else:
                            patch -= 1
                    else:
                        raise ValueError(
                            "Invalid change parameter. Use either 'increment' or 'decrement'."
                        )
                elif version_part == "minor":
                    if change == "increment":
                        minor = (minor + 1) % (max_minor + 1)
                        major += minor == 0
                    elif change == "decrement":
                        minor = (minor - 1) % (max_minor + 1)
                    else:
                        raise ValueError(
                            "Invalid change parameter. Use either 'increment' or 'decrement'."
                        )
                elif version_part == "major":
                    if change == "increment":
                        major += 1
                    elif change == "decrement":
                        major = max(0, major - 1)
                    else:
                        raise ValueError(
                            "Invalid change parameter. Use either 'increment' or 'decrement'."
                        )
                else:
                    raise ValueError(
                        "Invalid version_part parameter. Use either 'major', 'minor', or 'patch'."
                    )

                version = f"{major}.{minor}.{patch}"

                # Preserve the original format of the line
                new_line = re.sub(
                    version_regex,
                    lambda m: m.group(0).replace(
                        m.group(1) + "." + m.group(2) + "." + m.group(3), version
                    ),
                    line,
                )
                lines[i] = new_line
                break

        FileUtils.write_to_file(filepath, lines)
        if not version:
            print(f"Error: No version found in {filepath}")
        return version

    @staticmethod
    def update_requirements(file_path=None, inc=None, exc=None) -> list:
        """Update the requirements.txt file with current versions of packages.

        Args:
            file_path: Path to the requirements.txt file. Defaults to the caller's directory.
                      If a relative path is given, it's relative to the caller's directory.
            inc: Patterns or objects to include in the update.
            exc: Patterns or objects to exclude from the update.

        Returns:
            list: Updated requirements with their versions.
                 Example: ['package1==1.0.0', 'package2==2.3.4']
        """
        import os
        import inspect
        import pkg_resources
        from pythontk.iter_utils import IterUtils

        # Determine the caller's directory
        caller_frame = inspect.stack()[1]
        caller_path = caller_frame.filename
        caller_dir = os.path.dirname(caller_path)

        if file_path is None:
            file_path = os.path.join(caller_dir, "requirements.txt")
        else:
            file_path = os.path.abspath(os.path.join(caller_dir, file_path))

        try:
            with open(file_path, "r") as file:
                lines = file.readlines()

            updated_lines = []
            for line in lines:
                if line.strip() and not line.startswith("#"):
                    package_name = line.strip().split("==")[0]
                    if package_name in IterUtils.filter_list([package_name], inc, exc):
                        try:
                            version = pkg_resources.get_distribution(
                                package_name
                            ).version
                            updated_lines.append(f"{package_name}=={version}\n")
                        except Exception as e:
                            print(f"Error updating version for {package_name}: {e}")
                    else:
                        updated_lines.append(line)
                else:
                    updated_lines.append(line)

            with open(file_path, "w") as file:
                file.writelines(updated_lines)

        except FileNotFoundError:
            print(f"File not found: {file_path}")

        return [
            line.strip()
            for line in updated_lines
            if line.strip() and not line.startswith("#")
        ]


class PkgManager(PkgManagerHelperMixin, help_mixin.HelpMixin):
    """A class that encapsulates package management functionalities using pip.

    This class combines the capabilities of PkgManagerHelperMixin and HelpMixin to
    offer an integrated environment for managing Python packages and accessing help
    documentation for different methods. It provides an interface to execute pip commands
    and parse their output, along with handling Python environment-specific information.

    Methods:
        pip: Execute a pip command and handle its output.
        _get_startupinfo: Prepare startup information for subprocesses on Windows.
        _parse_command: Convert a pip command string into a list format.
        _process_output: Process and format the output from pip commands.
        _is_informational_message: Check if a stderr message is informational.
        _convert_output: Convert command output into a structured format.
        _is_list_format: Determine if the output is in a list format.
        _parse_list_format: Parse list format output into a dictionary.
        _is_key_value_format: Determine if the output is in key-value pair format.
        _parse_key_value_format: Parse key-value pair output into a dictionary.

    The PkgManager class serves as a comprehensive tool for package management in Python,
    supporting a range of pip functionalities and enhancing user experience with formatted
    help messages and command outputs.
    """

    def __init__(self, python_path=sys.executable):
        self.python_path = python_path

    def _get_startupinfo(self):
        """Prepare startup information for subprocess on Windows."""
        if sys.platform == "win32":
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            startupinfo.wShowWindow = subprocess.SW_HIDE
            return startupinfo
        return None

    def pip(self, command, output_as_string=False):
        """Execute a pip command and return the output."""
        full_command = [self.python_path, "-m", "pip"] + self._parse_command(command)
        startupinfo = self._get_startupinfo()

        try:
            result = subprocess.run(
                full_command,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                startupinfo=startupinfo,
            )
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Command '{' '.join(full_command)}' failed: {e.stderr}")

        return self._process_output(result.stdout, result.stderr, output_as_string)

    def _parse_command(self, command):
        """Parse the pip command from string to list format."""
        if isinstance(command, str):
            return re.findall(r'(?:[^\s,"]|"(?:\\.|[^"])*")+', command)
        return command

    def _process_output(self, stdout, stderr, output_as_string):
        """Process the output from the pip command."""
        if self._is_informational_message(stderr):
            stderr = ""

        output = stdout if not stderr else stdout + "\nError:\n" + stderr
        return output.strip() if output_as_string else self._convert_output(output)

    def _is_informational_message(self, stderr):
        """Check if the stderr message is informational."""
        informational_patterns = ["A new release of pip available"]
        return any(pattern in stderr for pattern in informational_patterns)

    def _convert_output(self, output):
        """Convert the output into a dictionary, list, or return as string."""
        try:
            return json.loads(output)
        except json.JSONDecodeError:
            lines = output.split("\n")
            if self._is_list_format(lines):
                return self._parse_list_format(lines)
            elif self._is_key_value_format(lines):
                return self._parse_key_value_format(lines)
            return output

    def _is_list_format(self, lines):
        """Determine if the output is in 'pip list' table format."""
        return all(len(line.split()) >= 2 for line in lines if line)

    def _parse_list_format(self, lines):
        """Parse 'pip list' table format into a dictionary."""
        output_dict = {}
        data_start = False  # Flag to indicate the start of data lines
        for line in lines:
            if "---" in line:  # Detect the line of dashes
                data_start = True
                continue
            if data_start and line.strip():
                package, version = line.split(maxsplit=1)
                output_dict[package] = version
        return output_dict

    def _is_key_value_format(self, lines):
        """Determine if the output is in key-value pair format."""
        return any(":" in line for line in lines if line)

    def _parse_key_value_format(self, lines):
        """Parse key-value pair format into a dictionary."""
        output_dict = {}
        for line in lines:
            if ":" in line:
                key, value = line.split(":", 1)
                output_dict[key.strip().lower()] = value.strip()
        return output_dict


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    pkg_mgr = PkgManager("C:/Program Files/Autodesk/Maya2025/bin/mayapy.exe")
    # output = pkg_mgr.get_installed_packages("")

    # Test various pip commands
    output = pkg_mgr.pip("list", output_as_string=1)
    # print(output)

    output = pkg_mgr.pip("show pythontk")
    print(output["version"])

    # output = pkg_mgr.pip("freeze")
    # print(output)

    # output = pkg_mgr.pip("check")
    # print(output)

    print("Installed Packages:")
    print(pkg_mgr.list_packages())

    # Show details of a specific package (e.g., 'numpy')
    print("\nPackage Details (numpy):")
    print(pkg_mgr.package_details("numpy"))

    # List outdated packages
    print("\nOutdated Packages:")
    print(pkg_mgr.list_outdated_packages())

    # Check if a specific package (e.g., 'numpy') is outdated
    print("\nIs 'numpy' outdated?")
    print(pkg_mgr.is_outdated("numpy"))

    pkg_mgr.help()

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
from typing import List, Dict, Any, Optional
import json


class HierarchyDiff:
    """Generic data class to hold hierarchical difference results.

    This class can be used for any hierarchical structure comparison,
    not just Maya scenes. Examples include file systems, XML/JSON structures,
    organizational charts, etc.
    """

    def __init__(self):
        """Initialize empty diff result."""
        self.missing: List[str] = []
        self.extra: List[str] = []
        self.renamed: List[str] = []
        self.reparented: List[str] = []
        self.fuzzy_matches: List[Dict[str, str]] = []
        self.metadata: Dict[str, Any] = {}

    def is_valid(self) -> bool:
        """Check if hierarchy has no significant differences.

        Returns:
            True if no missing, renamed, or reparented items exist
        """
        return not (self.missing or self.renamed or self.reparented)

    def has_differences(self) -> bool:
        """Check if any differences exist (including extra items).

        Returns:
            True if any differences exist
        """
        return bool(self.missing or self.extra or self.renamed or self.reparented)

    def total_issues(self) -> int:
        """Get total count of all issues.

        Returns:
            Sum of all difference counts
        """
        return (
            len(self.missing)
            + len(self.extra)
            + len(self.renamed)
            + len(self.reparented)
        )

    def as_dict(self) -> Dict[str, List[str]]:
        """Convert to dictionary representation.

        Returns:
            Dictionary with difference lists
        """
        return {
            "missing": self.missing,
            "extra": self.extra,
            "renamed": self.renamed,
            "reparented": self.reparented,
            "fuzzy_matches": self.fuzzy_matches,
            "metadata": self.metadata,
        }

    def as_json(self, indent: Optional[int] = 2) -> str:
        """Convert to JSON string.

        Parameters:
            indent: JSON indentation level

        Returns:
            JSON string representation
        """
        return json.dumps(self.as_dict(), indent=indent)

    def save_to_file(self, filepath: str, indent: Optional[int] = 2) -> None:
        """Save diff result to JSON file.

        Parameters:
            filepath: Path to save the JSON file
            indent: JSON indentation level
        """
        with open(filepath, "w", encoding="utf-8") as f:
            json.dump(self.as_dict(), f, indent=indent)

    @classmethod
    def load_from_file(cls, filepath: str) -> "HierarchyDiff":
        """Load diff result from JSON file.

        Parameters:
            filepath: Path to the JSON file

        Returns:
            HierarchyDiff instance
        """
        with open(filepath, "r", encoding="utf-8") as f:
            data = json.load(f)

        result = cls()
        result.missing = data.get("missing", [])
        result.extra = data.get("extra", [])
        result.renamed = data.get("renamed", [])
        result.reparented = data.get("reparented", [])
        result.fuzzy_matches = data.get("fuzzy_matches", [])
        result.metadata = data.get("metadata", {})

        return result

    def clear(self) -> None:
        """Clear all diff results."""
        self.missing.clear()
        self.extra.clear()
        self.renamed.clear()
        self.reparented.clear()
        self.fuzzy_matches.clear()
        self.metadata.clear()

    def merge(self, other: "HierarchyDiff") -> None:
        """Merge another diff result into this one.

        Parameters:
            other: Another HierarchyDiff to merge
        """
        self.missing.extend(other.missing)
        self.extra.extend(other.extra)
        self.renamed.extend(other.renamed)
        self.reparented.extend(other.reparented)
        self.fuzzy_matches.extend(other.fuzzy_matches)
        self.metadata.update(other.metadata)

    def get_summary(self) -> Dict[str, int]:
        """Get summary counts of all difference types.

        Returns:
            Dictionary with counts for each difference type
        """
        return {
            "missing": len(self.missing),
            "extra": len(self.extra),
            "renamed": len(self.renamed),
            "reparented": len(self.reparented),
            "fuzzy_matches": len(self.fuzzy_matches),
            "total_issues": self.total_issues(),
        }

    def filter_by_pattern(self, pattern: str, field: str = "missing") -> List[str]:
        """Filter items by regex pattern.

        Parameters:
            pattern: Regex pattern to match
            field: Field to filter ("missing", "extra", etc.)

        Returns:
            List of items matching the pattern
        """
        import re

        field_data = getattr(self, field, [])
        if isinstance(field_data, list):
            return [item for item in field_data if re.search(pattern, item)]
        return []

    def add_metadata(self, key: str, value: Any) -> None:
        """Add metadata to the diff result.

        Parameters:
            key: Metadata key
            value: Metadata value
        """
        self.metadata[key] = value

    def __str__(self) -> str:
        """String representation of diff result."""
        summary = self.get_summary()
        return (
            f"HierarchyDiff(missing={summary['missing']}, "
            f"extra={summary['extra']}, renamed={summary['renamed']}, "
            f"reparented={summary['reparented']})"
        )

    def __repr__(self) -> str:
        """Detailed string representation."""
        return self.__str__()


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    # Example usage
    print("=== HierarchyDiff Example ===")

    # Create a diff result
    diff = HierarchyDiff()
    diff.missing = ["item1", "item2"]
    diff.extra = ["item3"]
    diff.renamed = ["item4"]
    diff.add_metadata("analysis_time", "2024-08-02T10:30:00")
    diff.add_metadata("source", "test_hierarchy")

    print(f"Diff summary: {diff}")
    print(f"Is valid: {diff.is_valid()}")
    print(f"Has differences: {diff.has_differences()}")
    print(f"Total issues: {diff.total_issues()}")

    # JSON export/import
    json_str = diff.as_json()
    print(f"\nJSON representation:\n{json_str}")

    # Save and load example
    # diff.save_to_file("hierarchy_diff.json")
    # loaded_diff = HierarchyDiff.load_from_file("hierarchy_diff.json")

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------
# This module provides a general-purpose data structure for hierarchy difference analysis.
# It can be used for:
#
# - File system comparison
# - Database schema comparison
# - XML/JSON structure comparison
# - Maya scene hierarchy comparison
# - Any hierarchical data structure analysis
#
# The class includes utilities for JSON serialization, filtering, and metadata management.
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import inspect
from typing import get_type_hints
import textwrap


class HelpMixin:
    """A mixin class for providing dynamic help information based on different docstring formats.

    This mixin class facilitates the generation and display of formatted help information
    for methods within a class. It supports multiple established docstring formats including
    Google Style, NumPy/SciPy Style, reStructuredText, as well as a custom format. The class
    provides methods to detect the docstring format, format the docstring for display, and
    output help information for either a specific method or all methods within a class.

    Attributes:
        FORMATS (dict): A dictionary defining the structure for known docstring formats.
                        Each format is identified by a key, and the value is a list of
                        recognizable section headers for that format.
    Methods:
        help: Display help information for the current class or a specific method.
        _detect_format: Analyze a docstring to detect its format based on predefined templates.
        _format_docstring: Format a given docstring based on its detected format.

    The `help` method can be used to print formatted help information for any method
    within the class or the class itself, providing a quick and convenient way to access
    documentation. The mixin is designed to be flexible and adaptable to different
    documentation styles, making it a versatile tool for enhancing code readability and maintainability.
    """

    # Define the structure for known docstring formats
    FORMATS = {
        "google": [
            "Args:",
            "Returns:",
            "Raises:",
        ],
        "numpy": [
            "Parameters",
            "Returns",
            "Raises",
            "See Also",
            "Notes",
            "References",
            "Examples",
        ],
        "restructuredtext": [
            ":param",
            ":returns:",
            ":raises:",
        ],
        "custom": [
            "Parameters:",
            "Properties:",
            "Methods:",
            "Attributes:",
            "Example:",
            "Returns:",
            "Raises:",
            "Notes:",
            "Usage:",
        ],
    }

    @classmethod
    def help(cls, method_name=None):
        """Display help information for the current class or a specific method.

        Parameters:
            method_name: Optional. The name of the method to get help for.
        """
        entity = cls
        help_msg = f"Class: {cls.__name__}\nInherits from: {', '.join([base.__name__ for base in cls.__bases__])}\n\n"
        help_msg += f"Available Methods (use '{cls.__name__}.help('method_name')' for full docstring with examples):\n"
        print(help_msg)

        if method_name:
            method = getattr(entity, method_name, None)
            if method and inspect.isfunction(method):
                cls._display_method_help(method, method_name)
            else:
                print(f"No help available for method '{method_name}'.")
        else:
            # Display general help for methods unique to Switchboard class
            for attr_name in dir(entity):
                if attr_name.startswith("_") or hasattr(super(cls, entity), attr_name):
                    continue  # Skip private attributes and inherited methods

                attr = getattr(entity, attr_name)
                if callable(attr):
                    try:
                        signature = inspect.signature(attr)
                        method_sig = f"{attr_name}{signature}"
                        first_line_doc = ""
                        if attr.__doc__:
                            first_line_doc = (
                                textwrap.dedent(attr.__doc__).split("\n")[0].strip()
                            )
                        print(f"  - {method_sig}:\n\t{first_line_doc}\n")
                    except (TypeError, ValueError):
                        print(f"  - {attr_name}: (Method with complex signature)\n")

    @classmethod
    def _display_method_help(cls, method, method_name):
        """Display detailed help for a specific method."""
        doc = textwrap.dedent(method.__doc__).strip() if method.__doc__ else ""
        formatted_doc = cls._format_docstring(doc)

        # Get method signature and type hints
        signature = inspect.signature(method)
        type_hints = get_type_hints(method)

        print(f"Help for method '{method_name}':\n{formatted_doc}")
        print(f"\nSignature: {method_name}{signature}")

        # Display argument types and default values
        for param_name, param in signature.parameters.items():
            param_type = type_hints.get(param_name, "Any")
            default = (
                f" = {param.default}"
                if param.default is not inspect.Parameter.empty
                else ""
            )
            print(f"    {param_name}: {param_type}{default}")

        # Display return type
        return_type = type_hints.get("return", "None")
        print(f"Returns: {return_type}")

    @classmethod
    def _detect_format(cls, docstring):
        """Detect the docstring format.

        Parameters:
            docstring: The original docstring to analyze.

        Returns:
            The name of the detected format.
        """
        for format_name, headers in cls.FORMATS.items():
            if any(header in docstring for header in headers):
                return format_name
        return "custom"

    @classmethod
    def _format_docstring(cls, docstring):
        """Format the docstring based on its detected format.

        Parameters:
            docstring: The original docstring to format.
        """
        format_name = cls._detect_format(docstring)
        headers = cls.FORMATS[format_name]
        formatted_doc = ["\nDescription:"]
        lines = docstring.split("\n")

        section_header = None
        for line in lines:
            line = line.strip()
            if line in headers:
                if section_header:
                    formatted_doc.append("\n")  # Divider between sections
                section_header = line
                formatted_doc.append(f"{line}")
            elif section_header:
                formatted_doc.append(f"    {line}")
            else:
                formatted_doc.append(f"    {line}")

        return "\n".join(formatted_doc)


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import threading
import logging as internal_logging
from typing import Union, List, Optional
from pythontk.core_utils import ClassProperty


class LoggerExt:
    _text_handler = None  # Can be instance or class

    # Define custom log levels
    SUCCESS = 25
    RESULT = 35
    NOTICE = 45

    # Base log formats
    BASE_FORMATS = {
        "default": "[%(levelname)s] %(name)s: %(message)s",
        "debug": "[%(levelname)s] %(name)s: %(message)s",
        "success": "[%(levelname)s] %(message)s",
        "result": "[%(levelname)s] %(message)s",
        "notice": "[%(levelname)s] %(message)s",
    }

    @classmethod
    def patch(cls, logger: internal_logging.Logger) -> None:
        """Patch the logger with additional methods and setup."""
        if getattr(logger, "_logger_ext_patched", False):
            return
        logger._logger_ext_patched = True

        # Preserve the original setLevel method
        if not hasattr(logger, "internal_setLevel"):
            logger.internal_setLevel = logger.setLevel

        # Register custom levels
        cls._register_custom_levels()

        # Initialize error spam prevention
        logger._error_cache = {}
        logger._spam_prevention_enabled = True
        logger._cache_duration = 300  # 5 minutes default

        # Patch logger methods
        cls._patch_logger_methods(logger)

        # Initialize log formats and formatter selector
        logger._formatter_selector = cls._select_formatter.__get__(logger)
        logger.set_log_prefix = cls._set_log_prefix.__get__(logger)
        logger.set_log_suffix = cls._set_log_suffix.__get__(logger)
        logger._log_prefix = ""
        logger._log_suffix = ""
        logger._log_timestamp = None  # "%H:%M:%S" example of time only

        # Add property for log_timestamp that updates formatters
        def get_log_timestamp(self):
            return getattr(self, "_log_timestamp", None)

        def set_log_timestamp(self, value):
            self._log_timestamp = value
            LoggerExt._update_handler_formatters(self)

        logger.__class__.log_timestamp = property(get_log_timestamp, set_log_timestamp)

        # Add default handlers if none exist
        if not logger.handlers:
            cls._add_handler(logger, handler_type="stream")

    @staticmethod
    def _register_custom_levels() -> None:
        """Register custom log levels."""
        levels = {
            LoggerExt.SUCCESS: "SUCCESS",
            LoggerExt.RESULT: "RESULT",
            LoggerExt.NOTICE: "NOTICE",
        }
        for level, name in levels.items():
            internal_logging.addLevelName(level, name)

    @staticmethod
    def _patch_logger_methods(logger: internal_logging.Logger) -> None:
        """Patch logger with additional methods."""
        # Handle methods that don't need self (take no args or non-self first arg)
        direct_methods = {
            "set_text_handler": LoggerExt._set_text_handler,
            "get_text_handler": LoggerExt._get_text_handler,
        }
        for name, method in direct_methods.items():
            setattr(logger, name, method)

        # Handle methods that need self as first argument
        def make_method_wrapper(method):
            def wrapper(*args, **kwargs):
                return method(logger, *args, **kwargs)

            return wrapper

        wrapped_methods = {
            "setLevel": LoggerExt._set_level,
            "add_file_handler": LoggerExt._add_file_handler,
            "add_stream_handler": LoggerExt._add_stream_handler,
            "add_text_widget_handler": LoggerExt._add_text_widget_handler,
            "setup_logging_redirect": LoggerExt._setup_logging_redirect,
            "success": LoggerExt._success,
            "result": LoggerExt._result,
            "notice": LoggerExt._notice,
            "log_box": LoggerExt._log_box,
            "log_divider": LoggerExt._log_divider,
            "log_raw": LoggerExt._log_raw,
            "hide_logger_name": LoggerExt._hide_logger_name,
            "error_once": LoggerExt._error_once,
            "warning_once": LoggerExt._warning_once,
            "set_spam_prevention": LoggerExt._set_spam_prevention,
            "clear_error_cache": LoggerExt._clear_error_cache,
        }

        for name, method in wrapped_methods.items():
            setattr(logger, name, make_method_wrapper(method))

    @staticmethod
    def _select_formatter(self, level: int) -> internal_logging.Formatter:
        """Select formatter based on the log level."""
        prefix = getattr(self, "_log_prefix", "")
        suffix = getattr(self, "_log_suffix", "")
        base_format = LoggerExt._get_base_format(
            level, logger=self
        )  # Pass self as logger here
        format_string = base_format.replace(
            "%(message)s", f"{prefix}%(message)s{suffix}"
        )
        log_timestamp = getattr(self, "_log_timestamp", None)
        if log_timestamp:
            format_string = "[%(asctime)s] " + format_string
            formatter = internal_logging.Formatter(format_string, datefmt=log_timestamp)
        else:
            formatter = internal_logging.Formatter(format_string)
        return formatter

    @staticmethod
    def _get_base_format(level: int, logger=None) -> str:
        """Return the base format string based on the log level."""
        # Get the original format based on level
        if level == LoggerExt.SUCCESS:
            fmt = LoggerExt.BASE_FORMATS["success"]
        elif level == LoggerExt.RESULT:
            fmt = LoggerExt.BASE_FORMATS["result"]
        elif level == LoggerExt.NOTICE:
            fmt = LoggerExt.BASE_FORMATS["notice"]
        elif level <= internal_logging.DEBUG:
            fmt = LoggerExt.BASE_FORMATS["debug"]
        else:
            fmt = LoggerExt.BASE_FORMATS["default"]

        # If we have a logger instance and it has _hide_logger_name set to False,
        # strip the name from the format
        if logger and hasattr(logger, "_hide_logger_name") and logger._hide_logger_name:
            fmt = fmt.replace("%(name)s: ", "")

        return fmt

    @staticmethod
    def _add_handler(
        logger: internal_logging.Logger, handler_type: str, **kwargs
    ) -> None:
        """Add a handler to the logger."""
        handler = None
        if handler_type == "stream":
            handler = internal_logging.StreamHandler()
        elif handler_type == "file":
            handler = internal_logging.FileHandler(
                kwargs.get("filename", "logfile.log")
            )
        elif handler_type == "text_widget":
            handler_cls = LoggerExt._get_text_handler()
            handler = handler_cls(kwargs.get("widget"))

        if handler:
            level = kwargs.get("level", internal_logging.WARNING)
            handler.setLevel(level)
            handler.setFormatter(logger._formatter_selector(level))
            logger.addHandler(handler)
            logger.debug(f"{handler_type.capitalize()} handler added")

    @staticmethod
    def _add_file_handler(
        self, filename: str = "logfile.log", level: int = internal_logging.WARNING
    ) -> None:
        """Add a file handler to the logger."""
        LoggerExt._add_handler(
            self, handler_type="file", filename=filename, level=level
        )

    @staticmethod
    def _add_stream_handler(self, level: int = internal_logging.WARNING) -> None:
        """Add a stream handler to the logger."""
        LoggerExt._add_handler(self, handler_type="stream", level=level)

    @staticmethod
    def _add_text_widget_handler(
        self, text_widget: object, level: int = internal_logging.WARNING
    ) -> None:
        """Add a text widget handler to the logger."""
        LoggerExt._add_handler(
            self, handler_type="text_widget", widget=text_widget, level=level
        )

    @staticmethod
    def _set_log_prefix(self, prefix: str) -> None:
        """Set a prefix that will appear before all log messages."""
        self._log_prefix = prefix
        LoggerExt._update_handler_formatters(self)

    @staticmethod
    def _set_log_suffix(self, suffix: str) -> None:
        """Set a suffix that will appear after all log messages."""
        self._log_suffix = suffix
        LoggerExt._update_handler_formatters(self)

    @staticmethod
    def _update_handler_formatters(logger: internal_logging.Logger) -> None:
        """Update all handler formatters with the current prefix/suffix."""
        for handler in logger.handlers:
            handler.setFormatter(logger._formatter_selector(handler.level))

    @staticmethod
    def _set_level(self, level: Union[int, str]) -> None:
        """Set the log level."""
        if isinstance(level, str):
            level = internal_logging._nameToLevel.get(
                level.upper(), internal_logging.INFO
            )
        self.internal_setLevel(level)  # Call the preserved original method

    @staticmethod
    def _set_text_handler(handler: Union[type, object]) -> None:
        """Set a custom text handler class or instance."""
        LoggerExt._text_handler = handler

    @staticmethod
    def _get_text_handler() -> type:
        if LoggerExt._text_handler is None:
            return DefaultTextLogHandler
        if isinstance(LoggerExt._text_handler, type):
            return LoggerExt._text_handler  # it's a class
        return LoggerExt._text_handler.__class__  # it's an instance

    @staticmethod
    def _log_raw(self, message: str) -> None:
        """Write a raw message without level, prefix, or formatting."""
        # Write directly to all handler streams (console, files)
        for handler in self.handlers:
            stream = getattr(handler, "stream", None)
            if stream:
                try:
                    stream.write(message + "\n")
                    stream.flush()
                except Exception as e:
                    print(f"Logging error (raw write): {e}")
            else:  # For handlers without a stream (e.g., DefaultTextLogHandler), use emit
                try:
                    record = self.makeRecord(
                        name=self.name,
                        level=internal_logging.INFO,
                        fn="",
                        lno=0,
                        msg=message,
                        args=None,
                        exc_info=None,
                    )
                    record.raw = True  # <-- ADD THIS
                    handler.emit(record)

                except Exception as e:
                    print(f"Logging error (raw emit): {e}")

    @staticmethod
    def _log_box(self, title: str, items: List[str] = None, align: str = "left") -> int:
        """Print an ASCII box with title and optional list of lines. Returns box width."""
        padding = 2
        content = [title] + (items or [])
        longest = max(len(line) for line in content)
        inner_width = longest + padding * 2
        width = inner_width + 2  # full box width including sides

        top = "╔" + "═" * inner_width + "╗"

        if align == "left":
            title_text = " " * padding + title.ljust(longest) + " " * padding
        elif align == "right":
            title_text = " " * padding + title.rjust(longest) + " " * padding
        else:  # center
            title_text = " " * padding + title.center(longest) + " " * padding

        mid = "║" + title_text + "║"
        sep = "╟" + "─" * inner_width + "╢"
        bottom = "╚" + "═" * inner_width + "╝"

        LoggerExt._log_raw(self, top)
        LoggerExt._log_raw(self, mid)

        if items:
            LoggerExt._log_raw(self, sep)
            for item in items:
                item_line = " " + item.ljust(inner_width - 1)
                LoggerExt._log_raw(self, f"║{item_line}║")

        LoggerExt._log_raw(self, bottom)

        return width

    @staticmethod
    def _log_divider(self, width: Optional[int] = None, char: str = "─") -> None:
        """Print a clean divider line. If width is given, use that."""
        if width is None:
            width = 60  # fallback default
        LoggerExt._log_raw(self, char * width)

    # Public API for the custom log levels (SUCCESS, RESULT, NOTICE)
    @staticmethod
    def _success(self, msg: str, *args, **kwargs) -> None:
        # Call the original log method to avoid recursion
        internal_logging.Logger.__dict__["log"](
            self, LoggerExt.SUCCESS, f"{msg}", *args, **kwargs
        )

    @staticmethod
    def _result(self, msg: str, *args, **kwargs) -> None:
        internal_logging.Logger.__dict__["log"](
            self, LoggerExt.RESULT, f"{msg}", *args, **kwargs
        )

    @staticmethod
    def _notice(self, msg: str, *args, **kwargs) -> None:
        internal_logging.Logger.__dict__["log"](
            self, LoggerExt.NOTICE, f"{msg}", *args, **kwargs
        )

    @staticmethod
    def _set_log_prefix(self, prefix: str) -> None:
        """Set a prefix that will appear before all log messages."""
        self._log_prefix = prefix
        # Update all existing handlers with the new prefix
        for handler in self.handlers:
            handler.setFormatter(self._formatter_selector(handler.level))

    @staticmethod
    def _set_log_suffix(self, suffix: str) -> None:
        """Set a suffix that will appear after all log messages."""
        self._log_suffix = suffix
        # Update all existing handlers with the new suffix
        for handler in self.handlers:
            handler.setFormatter(self._formatter_selector(handler.level))

    @staticmethod
    def _setup_logging_redirect(
        self, target: Union[str, object], level: int = internal_logging.INFO
    ) -> None:
        """Redirect logging output to a specified target.
        :param target: Can be a filename (str), a stream (e.g., sys.stdout), or a widget (object).
        :param level: The log level for the redirection.
        """
        self.setLevel(level)  # <-- Always set logger level!
        if isinstance(target, str):
            self.add_file_handler(filename=target, level=level)
        elif hasattr(target, "write"):
            stream_handler = internal_logging.StreamHandler(stream=target)
            stream_handler.setLevel(level)
            stream_handler.setFormatter(self._formatter_selector(level))
            self.addHandler(stream_handler)
        elif hasattr(target, "append"):
            self.add_text_widget_handler(text_widget=target, level=level)
        else:
            raise ValueError("Unsupported target type for logging redirection.")

    @staticmethod
    def _hide_logger_name(self, show: bool = False) -> None:
        """Control whether the logger name is displayed in log messages.

        Args:
            show: If True (default), include the logger name in messages.
                  If False, omit the name portion for cleaner output.
        """
        self._hide_logger_name = show
        LoggerExt._update_handler_formatters(self)

    @staticmethod
    def _should_log_error(
        logger: internal_logging.Logger, message: str, cache_key: Optional[str] = None
    ) -> tuple[bool, str]:
        """Check if an error should be logged to prevent spam."""
        if not getattr(logger, "_spam_prevention_enabled", True):
            return True, ""

        import time

        # Generate cache key if not provided
        if cache_key is None:
            import hashlib

            cache_key = hashlib.md5(message.encode()).hexdigest()[:12]

        current_time = time.time()
        cache_duration = getattr(logger, "_cache_duration", 300)
        error_cache = getattr(logger, "_error_cache", {})

        # Check if we've seen this error recently
        if cache_key in error_cache:
            last_time, count = error_cache[cache_key]

            # If within cache duration, increment count and skip logging
            if current_time - last_time < cache_duration:
                error_cache[cache_key] = (last_time, count + 1)
                return (
                    False,
                    f" (suppressed {count} similar error{'s' if count != 1 else ''})",
                )

        # Log the error and cache it
        error_cache[cache_key] = (current_time, 1)
        return True, ""

    @staticmethod
    def _error_once(
        logger: internal_logging.Logger,
        message: str,
        *args,
        cache_key: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Log an error with automatic spam prevention."""
        should_log, suffix = LoggerExt._should_log_error(logger, message, cache_key)

        if should_log:
            # Log the full error with suffix
            full_message = f"{message}{suffix}"
            logger.error(full_message, *args, **kwargs)
        else:
            # Log a brief debug message for suppressed errors
            logger.debug(f"Suppressed duplicate error: {message[:50]}...")

    @staticmethod
    def _warning_once(
        logger: internal_logging.Logger,
        message: str,
        *args,
        cache_key: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Log a warning with automatic spam prevention."""
        should_log, suffix = LoggerExt._should_log_error(logger, message, cache_key)

        if should_log:
            full_message = f"{message}{suffix}"
            logger.warning(full_message, *args, **kwargs)
        else:
            logger.debug(f"Suppressed duplicate warning: {message[:50]}...")

    @staticmethod
    def _set_spam_prevention(
        logger: internal_logging.Logger, enabled: bool = True, cache_duration: int = 300
    ) -> None:
        """Configure spam prevention settings."""
        logger._spam_prevention_enabled = enabled
        logger._cache_duration = cache_duration
        if not enabled:
            logger._error_cache.clear()

    @staticmethod
    def _clear_error_cache(logger: internal_logging.Logger) -> None:
        """Clear the error cache."""
        logger._error_cache.clear()


class DefaultTextLogHandler(internal_logging.Handler):
    """A generic thread-safe logging handler that writes logs to any widget
    supporting .append(str). Supports raw output, optional HTML color formatting,
    and optional monospace font styling.
    """

    def __init__(self, widget: object, use_html: bool = True, monospace: bool = False):
        super().__init__()
        self.widget = widget
        self.setLevel(internal_logging.NOTSET)
        self.use_html = use_html
        self.monospace = monospace

    def emit(self, record: internal_logging.LogRecord) -> None:
        try:
            if getattr(record, "raw", False):
                msg = record.getMessage()
                threading.Timer(0, self._safe_append, args=(msg,)).start()
            else:
                msg = self.format(record)
                if self.use_html:
                    color = self.get_color(record.levelname)
                    formatted = f'<span style="color:{color}">{msg}</span>'
                    if self.monospace:
                        formatted = f'<pre style="margin:0">{formatted}</pre>'
                    threading.Timer(0, self._safe_append, args=(formatted,)).start()
                else:
                    threading.Timer(0, self._safe_append, args=(msg,)).start()
        except Exception as e:
            print(f"DefaultTextLogHandler emit error: {e}")

    def _safe_append(self, formatted_msg: str) -> None:
        try:
            if hasattr(self.widget, "append"):
                self.widget.append(formatted_msg)
            elif hasattr(self.widget, "insert"):
                self.widget.insert("end", formatted_msg + "\n")
            else:
                print(formatted_msg)
        except Exception as e:
            print(f"DefaultTextLogHandler append error: {e}")

    def get_color(self, level: str) -> str:
        return {
            "DEBUG": "#AAAAAA",  # Neutral gray
            "INFO": "#FFFFFF",  # Pure white
            "WARNING": "#FFF5B7",  # Pastel yellow
            "ERROR": "#FFCCCC",  # Pastel pink
            "CRITICAL": "#CC3333",  # Strong red
            "SUCCESS": "#CCFFCC",  # Pastel green
            "RESULT": "#CCFFFF",  # Pastel teal
            "NOTICE": "#E5CCFF",  # Pastel lavender
        }.get(
            level, "#FFFFFF"
        )  # fallback: pure white


class LoggingMixin:
    """Mixin class for logging utilities.

    Provides a logger for each class and a shared class logger across instances.
    Includes methods for setting log levels, adding handlers, and redirecting logs.
    """

    _logger: internal_logging.Logger = None
    _class_logger = None

    @ClassProperty
    def logger(cls) -> internal_logging.Logger:
        if cls.__dict__.get("_logger") is None:
            name = f"{cls.__module__}.{cls.__qualname__}"
            logger = internal_logging.Logger(name, internal_logging.NOTSET)  # CHANGED
            logger.propagate = False
            logger.parent = None
            LoggerExt.patch(logger)

            if not logger.handlers:
                logger.add_stream_handler()

            cls._logger = logger

        # sync all handler levels to match logger level
        for handler in cls._logger.handlers:
            handler.setLevel(cls._logger.level)

        return cls._logger

    @ClassProperty
    def class_logger(cls) -> internal_logging.Logger:
        if cls.__dict__.get("_class_logger") is None:
            name = f"{cls.__module__}.{cls.__name__}.class"
            logger = internal_logging.getLogger(name)
            logger.setLevel(internal_logging.NOTSET)  # CHANGED
            logger.propagate = False
            LoggerExt.patch(logger)
            cls._class_logger = logger
        return cls._class_logger

    @ClassProperty
    def logging(cls):
        """Access to Python's internal logging module (aliased)."""
        return internal_logging

    @classmethod
    def set_log_level(cls, level: int | str):
        """Set log level for the class logger and its handlers."""
        if isinstance(level, str):
            level = getattr(internal_logging, level.upper(), internal_logging.WARNING)

        cls.logger.setLevel(level)
        for handler in cls.logger.handlers:
            handler.setLevel(level)


# -------------------------------------------------------------------------------

if __name__ == "__main__":
    ...

# -------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
from typing import Dict, List, Any, Optional, Callable


class HierarchyIndexer:
    """Generic utilities for building and querying tree indices."""

    @staticmethod
    def _clean_namespace(name: str, separator: str = ":") -> str:
        """Remove namespace prefix from a name."""
        return name.split(separator)[-1] if separator in name else name

    @staticmethod
    def _split_hierarchy_path(path: str, separator: str = "|") -> List[str]:
        """Split a hierarchy path into components."""
        return path.split(separator) if path else []

    @staticmethod
    def _join_hierarchy_path(components: List[str], separator: str = "|") -> str:
        """Join path components into a hierarchy path."""
        return separator.join(components)

    @staticmethod
    def _clean_hierarchy_path(
        path: str, path_separator: str = "|", namespace_separator: str = ":"
    ) -> str:
        """Clean namespaces from all components of a hierarchy path."""
        components = HierarchyIndexer._split_hierarchy_path(path, path_separator)
        cleaned_components = [
            HierarchyIndexer._clean_namespace(comp, namespace_separator)
            for comp in components
        ]
        return HierarchyIndexer._join_hierarchy_path(cleaned_components, path_separator)

    @staticmethod
    def _normalize_path(
        path: str,
        clean_namespaces: bool = True,
        path_separator: str = "|",
        namespace_separator: str = ":",
    ) -> str:
        """Normalize a hierarchy path by cleaning and standardizing it."""
        if not path:
            return ""
        if clean_namespaces:
            return HierarchyIndexer._clean_hierarchy_path(
                path, path_separator, namespace_separator
            )
        else:
            return path

    @staticmethod
    def build_path_index(
        items: List[Any],
        get_path_func: Callable[[Any], str],
        path_separator: str = "|",
        clean_namespaces: bool = True,
        namespace_separator: str = ":",
    ) -> Dict[str, List[Any]]:
        """Build an index mapping cleaned paths to items.

        Args:
            items: List of tree items to index
            get_path_func: Function to extract path from an item
            path_separator: Character separating path components
            clean_namespaces: Whether to remove namespace prefixes
            namespace_separator: Character separating namespace from name

        Returns:
            Dictionary mapping normalized paths to lists of items
        """
        index = {}

        for item in items:
            raw_path = get_path_func(item)
            if raw_path:
                normalized_path = HierarchyIndexer._normalize_path(
                    raw_path, clean_namespaces, path_separator, namespace_separator
                )
                if normalized_path not in index:
                    index[normalized_path] = []
                index[normalized_path].append(item)

        return index

    @staticmethod
    def find_by_path(
        index: Dict[str, List[Any]],
        target_path: str,
        clean_namespaces: bool = True,
        path_separator: str = "|",
        namespace_separator: str = ":",
    ) -> List[Any]:
        """Find items in index by path.

        Args:
            index: Path index created by build_path_index
            target_path: Path to search for
            clean_namespaces: Whether to normalize the target path
            path_separator: Character separating path components
            namespace_separator: Character separating namespace from name

        Returns:
            List of items matching the path
        """
        normalized_path = HierarchyIndexer._normalize_path(
            target_path, clean_namespaces, path_separator, namespace_separator
        )
        return index.get(normalized_path, [])

    @staticmethod
    def find_by_tail_path(
        index: Dict[str, List[Any]],
        target_tail: str,
        num_components: int = 2,
        path_separator: str = "|",
    ) -> List[Any]:
        """Find items by matching the tail portion of their paths.

        Args:
            index: Path index created by build_path_index
            target_tail: Tail path to match
            num_components: Number of components in the tail
            path_separator: Character separating path components

        Returns:
            List of items whose paths end with the target tail
        """
        matches = []

        for path, items in index.items():
            components = HierarchyIndexer._split_hierarchy_path(path, path_separator)
            tail_components = components[-num_components:] if components else []
            path_tail = HierarchyIndexer._join_hierarchy_path(
                tail_components, path_separator
            )
            if path_tail == target_tail:
                matches.extend(items)

        return matches

    @staticmethod
    def get_path_components_index(
        items: List[Any], get_path_func: Callable[[Any], str], path_separator: str = "|"
    ) -> Dict[str, List[Any]]:
        """Build an index mapping individual path components to items.

        Args:
            items: List of tree items to index
            get_path_func: Function to extract path from an item
            path_separator: Character separating path components

        Returns:
            Dictionary mapping component names to lists of items
        """
        index = {}

        for item in items:
            path = get_path_func(item)
            if path:
                components = HierarchyIndexer._split_hierarchy_path(
                    path, path_separator
                )
                for component in components:
                    clean_component = HierarchyIndexer._clean_namespace(component)
                    if clean_component not in index:
                        index[clean_component] = []
                    if item not in index[clean_component]:
                        index[clean_component].append(item)

        return index

    @staticmethod
    def get_depth_index(
        items: List[Any], get_path_func: Callable[[Any], str], path_separator: str = "|"
    ) -> Dict[int, List[Any]]:
        """Build an index mapping path depths to items.

        Args:
            items: List of tree items to index
            get_path_func: Function to extract path from an item
            path_separator: Character separating path components

        Returns:
            Dictionary mapping depths (int) to lists of items
        """
        index = {}

        for item in items:
            path = get_path_func(item)
            if path:
                depth = len(
                    HierarchyIndexer._split_hierarchy_path(path, path_separator)
                )
                if depth not in index:
                    index[depth] = []
                index[depth].append(item)

        return index


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    ...

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
from typing import Any, Dict, List, Set, Tuple, Callable, Optional
from dataclasses import dataclass
from enum import Enum


class DifferenceType(Enum):
    """Types of differences that can be found between hierarchies."""

    MISSING = "missing"
    EXTRA = "extra"
    MODIFIED = "modified"
    MOVED = "moved"


@dataclass
class HierarchyDifference:
    """Represents a single difference between hierarchies."""

    type: DifferenceType
    path: str
    details: Dict[str, Any]
    item: Any = None

    def __str__(self) -> str:
        return f"{self.type.value}: {self.path} - {self.details}"


class HierarchyAnalyzer:
    """
    Analyzer for comparing hierarchical structures and identifying differences.

    This class provides comprehensive analysis capabilities for comparing
    two hierarchical structures and identifying various types of differences.
    """

    @staticmethod
    def compare_path_sets(
        current_paths: Set[str], reference_paths: Set[str], path_separator: str = "|"
    ) -> Dict[str, Set[str]]:
        """
        Compare two sets of hierarchical paths and categorize differences.

        Args:
            current_paths: Set of paths in current hierarchy
            reference_paths: Set of paths in reference hierarchy
            path_separator: Character separating path components

        Returns:
            Dictionary with 'missing', 'extra', and 'common' path sets
        """
        missing_paths = reference_paths - current_paths
        extra_paths = current_paths - reference_paths
        common_paths = current_paths & reference_paths

        return {"missing": missing_paths, "extra": extra_paths, "common": common_paths}

    @staticmethod
    def analyze_hierarchy_differences(
        current_items: List[Any],
        reference_items: List[Any],
        path_extractor: Callable[[Any], str],
        attribute_extractors: Dict[str, Callable[[Any], Any]] = None,
        path_separator: str = "|",
    ) -> List[HierarchyDifference]:
        """
        Perform comprehensive analysis of differences between hierarchies.

        Args:
            current_items: Items in current hierarchy
            reference_items: Items in reference hierarchy
            path_extractor: Function to extract path from item
            attribute_extractors: Functions to extract attributes for comparison
            path_separator: Character separating path components

        Returns:
            List of detected differences
        """
        if attribute_extractors is None:
            attribute_extractors = {}

        differences = []

        # Build path mappings
        current_map = {path_extractor(item): item for item in current_items}
        reference_map = {path_extractor(item): item for item in reference_items}

        current_paths = set(current_map.keys())
        reference_paths = set(reference_map.keys())

        # Basic path comparison
        path_comparison = HierarchyAnalyzer.compare_path_sets(
            current_paths, reference_paths, path_separator
        )

        # Missing items
        for missing_path in path_comparison["missing"]:
            differences.append(
                HierarchyDifference(
                    type=DifferenceType.MISSING,
                    path=missing_path,
                    details={"reason": "Path exists in reference but not in current"},
                    item=reference_map[missing_path],
                )
            )

        # Extra items
        for extra_path in path_comparison["extra"]:
            differences.append(
                HierarchyDifference(
                    type=DifferenceType.EXTRA,
                    path=extra_path,
                    details={"reason": "Path exists in current but not in reference"},
                    item=current_map[extra_path],
                )
            )

        # Compare attributes of common items
        for common_path in path_comparison["common"]:
            current_item = current_map[common_path]
            reference_item = reference_map[common_path]

            attribute_diffs = HierarchyAnalyzer._compare_item_attributes(
                current_item, reference_item, attribute_extractors
            )

            if attribute_diffs:
                differences.append(
                    HierarchyDifference(
                        type=DifferenceType.MODIFIED,
                        path=common_path,
                        details={"attribute_changes": attribute_diffs},
                        item=current_item,
                    )
                )

        return differences

    @staticmethod
    def _compare_item_attributes(
        current_item: Any,
        reference_item: Any,
        attribute_extractors: Dict[str, Callable[[Any], Any]],
    ) -> Dict[str, Dict[str, Any]]:
        """
        Compare attributes of two items.

        Args:
            current_item: Item from current hierarchy
            reference_item: Item from reference hierarchy
            attribute_extractors: Functions to extract attributes

        Returns:
            Dictionary of attribute differences
        """
        differences = {}

        for attr_name, extractor in attribute_extractors.items():
            try:
                current_value = extractor(current_item)
                reference_value = extractor(reference_item)

                if current_value != reference_value:
                    differences[attr_name] = {
                        "current": current_value,
                        "reference": reference_value,
                    }
            except Exception as e:
                differences[attr_name] = {
                    "error": f"Failed to compare attribute: {str(e)}"
                }

        return differences

    @staticmethod
    def detect_moved_items(
        differences: List[HierarchyDifference],
        similarity_threshold: float = 0.8,
        path_separator: str = "|",
    ) -> List[HierarchyDifference]:
        """
        Detect items that may have been moved rather than deleted/added.

        Args:
            differences: List of differences from analyze_hierarchy_differences
            similarity_threshold: Minimum similarity to consider items as moved
            path_separator: Character separating path components

        Returns:
            List of move differences detected
        """
        try:
            from difflib import SequenceMatcher
        except ImportError:
            # If difflib not available, return empty list
            return []

        move_differences = []

        missing_diffs = [d for d in differences if d.type == DifferenceType.MISSING]
        extra_diffs = [d for d in differences if d.type == DifferenceType.EXTRA]

        for missing_diff in missing_diffs[
            :
        ]:  # Use slice to allow removal during iteration
            missing_name = missing_diff.path.split(path_separator)[-1]
            best_match = None
            best_similarity = 0

            for extra_diff in extra_diffs[:]:
                extra_name = extra_diff.path.split(path_separator)[-1]
                similarity = SequenceMatcher(None, missing_name, extra_name).ratio()

                if similarity >= similarity_threshold and similarity > best_similarity:
                    best_match = extra_diff
                    best_similarity = similarity

            if best_match:
                # Create move difference
                move_diff = HierarchyDifference(
                    type=DifferenceType.MOVED,
                    path=missing_diff.path,
                    details={
                        "from_path": missing_diff.path,
                        "to_path": best_match.path,
                        "similarity": best_similarity,
                    },
                    item=missing_diff.item,
                )
                move_differences.append(move_diff)

                # Remove the matched differences from the lists
                if missing_diff in missing_diffs:
                    missing_diffs.remove(missing_diff)
                if best_match in extra_diffs:
                    extra_diffs.remove(best_match)

        return move_differences

    @staticmethod
    def categorize_differences(
        differences: List[HierarchyDifference], path_separator: str = "|"
    ) -> Dict[str, Dict[str, List[HierarchyDifference]]]:
        """
        Categorize differences by type and hierarchy level.

        Args:
            differences: List of differences to categorize
            path_separator: Character separating path components

        Returns:
            Nested dictionary organizing differences by type and level
        """
        categorized = {"by_type": {}, "by_level": {}, "by_root": {}}

        # Categorize by type
        for diff in differences:
            diff_type = diff.type.value
            if diff_type not in categorized["by_type"]:
                categorized["by_type"][diff_type] = []
            categorized["by_type"][diff_type].append(diff)

        # Categorize by hierarchy level
        for diff in differences:
            level = len(diff.path.split(path_separator)) if diff.path else 0
            if level not in categorized["by_level"]:
                categorized["by_level"][level] = []
            categorized["by_level"][level].append(diff)

        # Categorize by root component
        for diff in differences:
            root = diff.path.split(path_separator)[0] if diff.path else "unknown"
            if root not in categorized["by_root"]:
                categorized["by_root"][root] = []
            categorized["by_root"][root].append(diff)

        return categorized

    @staticmethod
    def generate_diff_report(
        differences: List[HierarchyDifference],
        include_details: bool = True,
        max_items_per_category: int = 10,
    ) -> str:
        """
        Generate a human-readable report of differences.

        Args:
            differences: List of differences to report
            include_details: Whether to include detailed information
            max_items_per_category: Maximum items to show per category

        Returns:
            Formatted difference report string
        """
        if not differences:
            return "No differences found between hierarchies."

        categorized = HierarchyAnalyzer.categorize_differences(differences)

        report_lines = [
            "=" * 60,
            "HIERARCHY DIFFERENCE ANALYSIS REPORT",
            "=" * 60,
            f"Total differences found: {len(differences)}",
            "",
        ]

        # Summary by type
        report_lines.append("SUMMARY BY TYPE:")
        for diff_type, type_diffs in categorized["by_type"].items():
            report_lines.append(f"  {diff_type.upper()}: {len(type_diffs)} items")
        report_lines.append("")

        # Detailed breakdown
        if include_details:
            for diff_type, type_diffs in categorized["by_type"].items():
                report_lines.append(f"{diff_type.upper()} ITEMS:")

                items_to_show = type_diffs[:max_items_per_category]
                for diff in items_to_show:
                    report_lines.append(f"  • {diff.path}")
                    if diff.details:
                        for key, value in diff.details.items():
                            report_lines.append(f"    {key}: {value}")

                if len(type_diffs) > max_items_per_category:
                    remaining = len(type_diffs) - max_items_per_category
                    report_lines.append(f"  ... and {remaining} more items")

                report_lines.append("")

        # Level distribution
        report_lines.append("DISTRIBUTION BY HIERARCHY LEVEL:")
        for level in sorted(categorized["by_level"].keys()):
            level_diffs = categorized["by_level"][level]
            report_lines.append(f"  Level {level}: {len(level_diffs)} items")

        report_lines.append("=" * 60)

        return "\n".join(report_lines)

    @staticmethod
    def export_differences_to_dict(
        differences: List[HierarchyDifference],
    ) -> Dict[str, Any]:
        """
        Export differences to a dictionary format for serialization.

        Args:
            differences: List of differences to export

        Returns:
            Dictionary representation of differences
        """
        return {
            "total_count": len(differences),
            "differences": [
                {"type": diff.type.value, "path": diff.path, "details": diff.details}
                for diff in differences
            ],
            "summary": {
                diff_type.value: len([d for d in differences if d.type == diff_type])
                for diff_type in DifferenceType
            },
        }

    @staticmethod
    def filter_differences(
        differences: List[HierarchyDifference],
        types: List[DifferenceType] = None,
        path_patterns: List[str] = None,
        exclude_patterns: List[str] = None,
    ) -> List[HierarchyDifference]:
        """
        Filter differences based on type and path patterns.

        Args:
            differences: List of differences to filter
            types: List of difference types to include
            path_patterns: List of path patterns to include (supports wildcards)
            exclude_patterns: List of path patterns to exclude

        Returns:
            Filtered list of differences
        """
        filtered = differences

        # Filter by type
        if types:
            filtered = [d for d in filtered if d.type in types]

        # Filter by path patterns
        if path_patterns:
            import fnmatch

            pattern_filtered = []
            for diff in filtered:
                for pattern in path_patterns:
                    if fnmatch.fnmatch(diff.path, pattern):
                        pattern_filtered.append(diff)
                        break
            filtered = pattern_filtered

        # Exclude patterns
        if exclude_patterns:
            import fnmatch

            exclude_filtered = []
            for diff in filtered:
                should_exclude = False
                for pattern in exclude_patterns:
                    if fnmatch.fnmatch(diff.path, pattern):
                        should_exclude = True
                        break
                if not should_exclude:
                    exclude_filtered.append(diff)
            filtered = exclude_filtered

        return filtered


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    ...

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# coding=utf-8
"""
Hierarchy utilities subpackage providing generic, reusable hierarchy matching and manipulation logic.

This subpackage contains framework-agnostic utilities for:
- Path manipulation and namespace handling
- Tree-like structure indexing and matching
- Hierarchy comparison and diff operations
- Widget-agnostic tree matching strategies

These utilities can be used across different GUI frameworks (Qt, Tkinter, etc.)
and different applications (Maya, Blender, etc.) without dependencies on specific frameworks.
"""

from .hierarchy_indexer import HierarchyIndexer
from .hierarchy_matching import HierarchyMatching
from .hierarchy_analyzer import HierarchyDifference, HierarchyAnalyzer

__all__ = [
    "HierarchyIndexer",
    "HierarchyMatching",
    "HierarchyDifference",
    "HierarchyAnalyzer",
]



import weakref
from functools import partial
from typing import Callable, Optional, Any, Type
from pythontk.core_utils import LoggingMixin


class Placeholder:
    def __init__(
        self,
        class_type: Type,
        factory: Optional[Callable] = None,
        *,
        args: Optional[tuple] = (),
        kwargs: Optional[dict] = None,
        meta: Optional[dict] = None,
    ):
        self.class_type = class_type
        self.factory = factory
        self.args = args or ()
        self.kwargs = kwargs or {}
        self.meta = meta or {}

    def info(self) -> dict:
        return {
            "type": self.class_type.__name__,
            "factory": self.factory.__name__ if self.factory else None,
            "args": self.args,
            "kwargs": self.kwargs,
            "meta": self.meta,
        }

    def create(self, *args, **kwargs):
        ctor = self.factory or self.class_type
        return ctor(*self.args, *args, **{**self.kwargs, **kwargs})

    def __repr__(self):
        return f"<Placeholder for {self.class_type.__name__}>"


class NamespaceHandler(LoggingMixin):
    """A NamespaceHandler that manages its own internal dictionary without attaching
    attributes directly to the owner object.

    Parameters:
        owner (Any): The owner object that the namespace is attached to.
        identifier (str, optional): An identifier for logging or tracking purposes.
        resolver (Callable[[str], Any], optional): A function that resolves attribute names.
        log_level (str, optional): The logging level for the logger.

    Example:
        def resolver(name: str) -> Any:
            # Custom resolver logic here
            return f"Resolved {name}" if name != "unknown_attr" else None

        handler = NamespaceHandler(owner=my_object, identifier="example", resolver=resolver)
        print(handler.some_attr)  # Calls the resolver if not cached
        handler.some_attr = "New Value"  # Sets the attribute directly
    """

    def __init__(
        self,
        owner: Any,
        identifier: str = None,
        resolver: Optional[Callable[[str], Any]] = None,
        use_weakref: bool = False,
        log_level: str = "WARNING",
    ):
        self.logger.setLevel(log_level)

        self.__dict__["Placeholder"] = Placeholder
        self.__dict__["_identifier"] = identifier
        self.__dict__["_resolver"] = resolver
        self.__dict__["_owner"] = owner
        self.__dict__["set"] = self.__setitem__
        self.__dict__["get"] = partial(
            lambda self, k, d=None, resolve_placeholders=True: self.resolve(
                k, d, resolve_placeholders
            ),
            self,
        )
        self.__dict__["raw"] = self.raw
        self.__dict__["_use_weakref"] = use_weakref
        if use_weakref:
            self.__dict__["_attributes"] = weakref.WeakValueDictionary()
        else:
            self.__dict__["_attributes"] = {}
        self.__dict__["_placeholders"] = {}

    @property
    def placeholders(self) -> dict[str, Any]:
        return self._placeholders

    def is_placeholder(self, value: Any) -> bool:
        return isinstance(value, self.Placeholder)

    def get_placeholder(self, key: str) -> Optional[Placeholder]:
        return self._placeholders.get(key)

    def set_placeholder(self, key: str, placeholder: Placeholder):
        self.logger.debug(f"[{self._identifier}] Set placeholder for: {key}")
        self._placeholders[key] = placeholder

    def resolve_all_placeholders(self):
        for key in list(self._placeholders):
            _ = self[key]  # Triggers resolution

    def has_placeholder(self, key: str) -> bool:
        return key in self._placeholders

    def __repr__(self):
        return (
            f"<NamespaceHandler id='{self._identifier}' keys={list(self.keys(True))}>"
        )

    def __contains__(self, key: str) -> bool:
        return key in self._attributes or key in self._placeholders

    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{name}'")

        # Check actual attributes
        if name in self._attributes:
            return self._attributes[name]

        # Check placeholders
        if name in self._placeholders:
            return self._placeholders[name]

        try:
            return self._resolve_and_cache(name)
        except KeyError:
            self.logger.debug(f"[{self._identifier}] Attribute '{name}' not found.")
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any):
        if name.startswith("_"):
            self.__dict__[name] = value
            return

        if isinstance(value, self.Placeholder):
            self._placeholders[name] = value
            self._attributes.pop(name, None)
            return

        self._placeholders.pop(name, None)

        if self._use_weakref:
            try:
                self._attributes[name] = value
            except TypeError:
                self._attributes.data[name] = value
        else:
            self._attributes[name] = value

    def __getitem__(self, key: str, resolve_placeholders: bool = True) -> Any:
        if key in self._attributes:
            return self._attributes[key]

        if key in self._placeholders:
            if not resolve_placeholders:
                return self._placeholders[key]
            return self._resolve_placeholder(key)

        return self._resolve_and_cache(key)

    def __setitem__(self, key: str, value: Any):
        if isinstance(value, self.Placeholder):
            self._placeholders[key] = value
            self._attributes.pop(key, None)  # Ensure not duplicated
            return

        self._placeholders.pop(key, None)  # Clean up any previous placeholder

        if self._use_weakref:
            try:
                self._attributes[key] = value
            except TypeError:
                self._attributes.data[key] = value
        else:
            self._attributes[key] = value

    def __delitem__(self, key: str):
        self._attributes.pop(key, None)
        self._placeholders.pop(key, None)

    def keys(self, inc_placeholders=False):
        keys = set(self._attributes.keys())
        if inc_placeholders:
            keys.update(self._placeholders.keys())
        return keys

    def items(self, inc_placeholders=False):
        combined = dict(self._attributes)
        if inc_placeholders:
            combined.update(self._placeholders)
        return combined.items()

    def values(self, inc_placeholders=False):
        values = list(self._attributes.values())
        if inc_placeholders:
            values += list(self._placeholders.values())
        return values

    def setdefault(self, key: str, default: Any = None) -> Any:
        if key in self:
            return self[key]
        self[key] = default
        return default

    def has(self, key: str) -> bool:
        return key in self._attributes

    def raw(self, key: str) -> Optional[Any]:
        if key in self._attributes:
            return self._attributes[key]
        if key in self._placeholders:
            return self._placeholders[key]
        return None

    def resolve(
        self, key: str, default: Any = None, resolve_placeholders: bool = True
    ) -> Any:
        try:
            return self.__getitem__(key, resolve_placeholders=resolve_placeholders)
        except KeyError:
            return default

    def is_resolving(self, key: str) -> bool:
        """Returns True if this key is currently being resolved (to prevent recursion)."""
        return getattr(self, f"__resolving__{key}", False)

    def _resolve_placeholder(self, key: str) -> Any:
        placeholder = self._placeholders[key]
        self.logger.debug(f"[{self._identifier}] Resolving placeholder: {key}")
        value = placeholder.create()
        self[key] = value
        del self._placeholders[key]
        return value

    def _resolve_and_cache(self, key: str) -> Any:
        attributes = self.__dict__["_attributes"]
        if key in attributes:
            return attributes[key]

        # Prevent recursion
        guard_key = f"__resolving__{key}"
        if getattr(self, guard_key, False):
            raise RuntimeError(f"Recursive resolution detected for key: '{key}'")
        setattr(self, guard_key, True)

        try:
            resolver = self.__dict__.get("_resolver")
            if resolver:
                resolved = resolver(key)
                if resolved is not None:
                    if self._use_weakref:
                        try:
                            attributes[key] = resolved
                        except TypeError:
                            attributes.data[key] = resolved
                            self.logger.debug(
                                f"[{self._identifier}] Not weakref-able, stored strong ref for: {key}"
                            )
                    else:
                        attributes[key] = resolved
                    return resolved
        finally:
            if hasattr(self, guard_key):
                delattr(self, guard_key)

        raise KeyError(key)


# --------------------------------------------------------------------------------------------


if __name__ == "__main__":
    ...


# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------
